import json
from functools import cache
from pathlib import Path
from typing import Optional

from pydantic import BaseModel


class DependencySpec(BaseModel):
    prefer_source: str
    repo: Optional[str] = None
    depends_on: Optional[list[str]] = []
    version: Optional[str] = None
    optional: bool = False


class AnalysisSpec(BaseModel):
    name: str
    header_version_key: Optional[str] = None
    requirements: dict[str, DependencySpec]


@cache
def get_specs() -> dict:
    dir = Path(__file__).parent
    files = dir.glob("*.json")
    specs = {}

    for file in files:
        try:
            name, spec = __load_spec(file)
        except (json.JSONDecodeError, ValueError):
            continue
        requirements = spec.pop("requirements")
        dep_specs = {
            name: DependencySpec(**spec_) for name, spec_ in requirements.items()
        }

        specs[name] = AnalysisSpec(**spec, requirements=dep_specs)

    return specs


def __load_spec(path: Path) -> tuple[str, dict]:
    with open(path, "r") as f:
        data = json.load(f)
    if "name" not in data:
        raise ValueError

    return data["name"], data


__all__ = ["specs"]



from enum import Enum
from typing import Callable

import astropy.units as u


class UnitConvention(Enum):
    COMOVING = "comoving"
    PHYSICAL = "physical"
    SCALEFREE = "scalefree"
    UNITLESS = "unitless"


ConventionConverters = dict[UnitConvention, Callable[[u.Quantity], u.Quantity]]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Optional

import astropy.cosmology.units as cu
import astropy.units as u
from astropy.constants import m_p

if TYPE_CHECKING:
    import h5py
    from astropy.cosmology import Cosmology
    from numpy.typing import ArrayLike

    from opencosmo.header import OpenCosmoHeader
from opencosmo.units.convention import UnitConvention
from opencosmo.units.converters import get_unit_transitions

KNOWN_UNITS = {
    "comoving Mpc/h": u.Mpc / cu.littleh,
    "comoving (Mpc/h)^2": (u.Mpc / cu.littleh) ** 2,
    "comoving km/s": u.km / u.s,
    "comoving (km/s)^2": (u.km / u.s) ** 2,
    "Msun/h": u.Msun / cu.littleh,
    "Msun/yr": u.Msun / u.yr,
    "K": u.K,
    "comoving (Msun/h * (km/s) * Mpc/h)": (u.Msun / cu.littleh)
    * (u.km / u.s)
    * (u.Mpc / cu.littleh),
    "log10(erg/s)": u.DexUnit("erg/s"),
    "h^2 keV / (comoving cm)^3": (cu.littleh**2) * u.keV / (u.cm**3),
    "keV * cm^2": u.keV * u.cm**2,
    "cm^-3": u.cm**-3,
    "Gyr": u.Gyr,
    "Msun/h / (comoving Mpc/h)^3": (u.Msun / cu.littleh) / (u.Mpc / cu.littleh) ** 3,
    "Msun/h * km/s": (u.Msun / cu.littleh) * (u.km / u.s),
    "H0^-1": (u.s * (1 * u.Mpc).to(u.km).value).to(u.year) / (100 * cu.littleh),
    "m_hydrogen": m_p,
    "Msun * (km/s)^2": (u.Msun) * (u.km / u.s) ** 2,
}


class UnitApplicator:
    def __init__(
        self,
        units: dict[UnitConvention, u.Unit],
        base_convention: UnitConvention,
        converters: dict[UnitConvention, Callable],
        invserse_converters: dict[UnitConvention, Callable],
    ):
        self.__units = units
        self.__base_convention = base_convention
        self.__converters = converters
        self.__inv_converters = invserse_converters

    @classmethod
    def static(cls, base_unit: u.Unit, base_convention: UnitConvention):
        """
        Unit applicator for a column that does not transform under changes in unit convention.
        This column CAN still be transformed explicitly to an equivalent unit.
        """
        units = {base_convention: base_unit}
        if base_convention in [UnitConvention.COMOVING, UnitConvention.SCALEFREE]:
            units[UnitConvention.PHYSICAL] = base_unit
        if base_convention == UnitConvention.SCALEFREE:
            units[UnitConvention.COMOVING] = base_unit
        return UnitApplicator(units, base_convention, {}, {})

    @classmethod
    def from_unit(
        cls,
        base_unit: u.Unit,
        base_convention: UnitConvention,
        cosmology: Cosmology,
        is_comoving: bool = True,
    ):
        if base_unit is None:
            # Certain "units" are not actually units (e.g. m_p)
            return UnitApplicator({}, base_convention, {}, {})

        if isinstance(base_unit, u.Quantity):
            trans, inv_trans, units = get_unit_transitions(
                base_unit.unit, base_convention, cosmology, is_comoving
            )
        else:
            trans, inv_trans, units = get_unit_transitions(
                base_unit, base_convention, cosmology, is_comoving
            )

        return UnitApplicator(units, base_convention, trans, inv_trans)

    @property
    def base_unit(self):
        return self.unit_in_convention(self.__base_convention)

    def unit_in_convention(self, convention: UnitConvention):
        return self.__units.get(convention)

    def apply(
        self,
        value: ArrayLike,
        convention: UnitConvention,
        convert_to: Optional[u.Unit] = None,
        unit_kwargs: dict[str, Any] = {},
    ) -> u.Quantity:
        if not self.__units or convention == UnitConvention.UNITLESS:
            return value
        if hasattr(value, "unit"):
            raise ValueError(
                "Units can only be applied to unitless scalars and numpy arrays"
            )
        new_value = value * self.__units[self.__base_convention]
        if convention != self.__base_convention:
            new_value = self.__convert(new_value, convention, unit_kwargs)

        if convert_to is not None:
            new_value = new_value.to(convert_to)

        return new_value

    def can_convert(self, to_: u.Unit, convention: UnitConvention):
        unit_to_convert = self.__units.get(convention)
        if unit_to_convert is None:
            return False
        return to_.is_equivalent(unit_to_convert)

    def convert_to_base(
        self,
        value: u.Quantity | float,
        convention: UnitConvention,
        unit_kwargs: dict[str, Any] = {},
    ):
        if not self.__units:
            return value

        if not isinstance(value, u.Quantity):
            value = value * self.__units[convention]

        converter = self.__inv_converters.get(convention)
        if converter is not None:
            return converter(value, **unit_kwargs)
        return value

    def __convert(
        self, value: u.Quantity, to_: UnitConvention, unit_kwargs: dict[str, Any]
    ) -> u.Quantity:
        converter = self.__converters.get(to_)
        if converter is not None:
            return converter(value, **unit_kwargs)
        return value


def get_unit_applicators_hdf5(
    group: h5py.Group, header: "OpenCosmoHeader", is_comoving: bool = True
):
    base_convention = UnitConvention(header.file.unit_convention)

    applicators = {}
    for name, column in group.items():
        base_unit = get_raw_units(column)
        applicators[name] = UnitApplicator.from_unit(
            base_unit, base_convention, header.cosmology, is_comoving
        )
    return applicators


def get_unit_applicators_dict(
    units: dict[str, u.Unit],
    base_convention: UnitConvention,
    cosmology: Cosmology,
    is_comoving: bool = True,
):
    applicators = {}
    for name, base_unit in units.items():
        applicators[name] = UnitApplicator.from_unit(
            base_unit, base_convention, cosmology, is_comoving
        )
    return applicators


def get_raw_units(column: h5py.Dataset):
    if "unit" in column.attrs:
        if (us := column.attrs["unit"]) == "None" or us == "":
            return None
        if (unit := KNOWN_UNITS.get(us)) is not None:
            return unit
        try:
            return u.Unit(us)
        except ValueError:
            return None

    return None



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Iterable, Mapping, Optional, Self

from opencosmo.collection import structure as sc
from opencosmo.dataset import Dataset
from opencosmo.io import io
from opencosmo.io.schemas import SimCollectionSchema

if TYPE_CHECKING:
    import astropy.units as u
    import h5py
    from astropy.cosmology import Cosmology

    from opencosmo.collection.protocols import Collection
    from opencosmo.dataset.column import ColumnMask
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.io.protocols import DataSchema
    from opencosmo.parameters import HaccSimulationParameters
    from opencosmo.spatial.protocols import Region


def verify_datasets_exist(file: h5py.File, datasets: Iterable[str]):
    """
    Verify a set of datasets exist in a given file.
    """
    if not set(datasets).issubset(set(file.keys())):
        raise ValueError(f"Some of {', '.join(datasets)} not found in file.")


class SimulationCollection(dict):
    """
    A collection of datasets of the same type from different
    simulations. In general this exposes the exact same API
    as the individual datasets, but maps the results across
    all of them.
    """

    def __init__(self, datasets: Mapping[str, Dataset | Collection]):
        self.update(datasets)

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        for dataset in self.values():
            try:
                dataset.close()
            except ValueError:
                continue

    def __repr__(self):
        n_collections = sum(
            1
            for v in self.values()
            if isinstance(v, (SimulationCollection, sc.StructureCollection))
        )
        n_datasets = sum(1 for v in self.values() if isinstance(v, Dataset))
        return (
            f"SimulationCollection({n_collections} collections, {n_datasets} datasets)"
        )

    @classmethod
    def open(cls, targets: list[io.OpenTarget], **kwargs) -> Collection | Dataset:
        targets_by_name = {
            target.group.name.split("/")[-1]: target for target in targets
        }
        if len(targets_by_name) != len(targets):
            raise ValueError(
                "Not all datasets in this SimulationCollection have unique names!"
            )

        datasets = {
            name: io.open_single_dataset(target)
            for name, target in targets_by_name.items()
        }

        if len(datasets) == 1:
            return next(iter(datasets.values()))
        return cls(datasets)

    def make_schema(self) -> DataSchema:
        schema = SimCollectionSchema()
        for name, dataset in self.items():
            ds_schema = dataset.make_schema()
            schema.add_child(ds_schema, name)

        return schema

    def __map(
        self,
        method,
        *args,
        construct=True,
        datasets: Optional[str | Iterable[str]] = None,
        **kwargs,
    ):
        """
        This type of collection will only ever be constructed if all the underlying
        datasets have the same data type, so it is always safe to map operations
        across all of them.
        """
        regular_kwargs = {}
        mapped_kwargs = {}
        if isinstance(datasets, str):
            datasets = [datasets]
        elif datasets is None:
            datasets = self.keys()
        requested_datasets = set(datasets)
        if not requested_datasets.issubset(self.keys()):
            raise ValueError(
                f"Unknown datasets {requested_datasets.difference(self.keys())}"
            )

        for name, value in kwargs.items():
            if isinstance(value, dict) and set(value.keys()) == requested_datasets:
                mapped_kwargs[name] = value
            else:
                regular_kwargs[name] = value

        output = {}
        for name in requested_datasets:
            dataset = self[name]
            dataset_mapped_kwargs = {key: kw[name] for key, kw in mapped_kwargs.items()}
            output[name] = getattr(dataset, method)(
                *args, **regular_kwargs, **dataset_mapped_kwargs
            )
        if construct:
            return SimulationCollection(output)
        return output

    def __map_attribute(self, attribute):
        return {k: getattr(v, attribute) for k, v in self.items()}

    @property
    def dtype(self) -> dict[str, str]:
        return {key: ds.header.file.data_dtype for key, ds in self.items()}

    @property
    def header(self) -> dict[str, OpenCosmoHeader]:
        return self.__map_attribute("header")

    @property
    def cosmology(self) -> dict[str, Cosmology]:
        """
        Get the cosmologies of the simulations in the collection

        Returns
        --------
        cosmologies: dict[str, astropy.cosmology.Cosmology]
        """
        return self.__map_attribute("cosmology")

    @property
    def redshift(self) -> dict[str, float | tuple[float, float]]:
        """
        Get the redshift slices or ranges for the simulations in the collection

        Returns
        --------
        redshifts: dict[str, float | tuple[float,float]]
        """
        return self.__map_attribute("redshift")

    @property
    def simulation(self) -> dict[str, HaccSimulationParameters]:
        """
        Get the simulation parameters for the simulations in the collection

        Returns
        --------
        simulation_parameters: dict[str, opencosmo.parameters.HaccSimulationParameters]
        """

        return self.__map_attribute("simulation")

    def bound(self, region: Region, select_by: Optional[str] = None) -> Self:
        """
        Restrict the datasets to some region. Note that the SimulationCollection does
        not do any checking to ensure its members have identical boxes. As a result
        this method can in principle fail for some of the simulations in the
        collection and not others. This should never happen when working with official
        OpenCosmo data products.

        See :doc:`spatial_ref` for details of how to construct regions.

        Parameters
        ----------
        region: opencosmo.spatial.Region
            The region to query

        Returns
        -------
        dataset: opencosmo.SimulationCollection
            The portion of each dataset inside the selected region

        """
        return self.__map("bound", region, select_by)

    def filter(self, *masks: ColumnMask, **kwargs) -> Self:
        """
        Filter the datasets in the collection. This method behaves
        exactly like :meth:`opencosmo.Dataset.filter` or
        :meth:`opencosmo.StructureCollection.filter`, but
        it applies the filter to all the datasets or collections
        within this collection. The result is a new collection.

        Parameters
        ----------
        filters:
            The filters constructed with :func:`opencosmo.col`

        Returns
        -------
        SimulationCollection
            A new collection with the same datasets, but only the
            particles that pass the filter.
        """
        return self.__map("filter", *masks, **kwargs)

    def select(self, *args, **kwargs) -> Self:
        """
        Select a set of columns in the datasets in this collection. This method
        calls the underlying method in :class:`opencosmo.Dataset`, or
        :class:`opencosmo.Collection` depending on the context. As such
        its behavior and arguments can vary depending on what this collection
        contains.

        Parameters
        ----------
        args:
            The arguments to pass to the select method. This is
            usually a list of column names to select.
        kwargs:
            The keyword arguments to pass to the select method.
            This is usually a dictionary of column names to select.

        Returns
        -------
        SimulationCollection
            A new collection with only the specified columns

        """
        return self.__map("select", *args, **kwargs)

    def drop(self, *args, **kwargs) -> Self:
        """
        Drop a set of columns from the datasets in the collection. This method
        calls the underlying method in :class:`opencosmo.Dataset`, or
        :class:`opencosmo.Collection` depending on the context. As such
        its behavior and arguments can vary depending on what this collection
        contains.

        Parameters
        ----------
        args:
            The arguments to pass to the select method. This is
            usually a list of column names to drop.
        kwargs:
            The keyword arguments to pass to the select method.
            This is usually a dictionary of column names to select.

        """
        return self.__map("drop", *args, **kwargs)

    def take(self, n: int, at: str = "random") -> Self:
        """
        Take a subest of rows from all datasets or collections in this collection.
        This method will delegate to the underlying method in
        :class:`opencosmo.Dataset`, or :class:`opencosmo.StructureCollection` depending
        on  the context. As such, behavior may vary depending on what this collection
        contains. See their documentation for more info.

        Parameters
        ----------
        n: int
            The number of rows to take
        at: str, default = "random"
            The method to use to take rows. Must be one of "start", "end", "random".

        """
        if any(len(ds) < n for ds in self.values()):
            raise ValueError(
                f"Not all datasets in this collection have at least {n} rows!"
            )
        return self.__map("take", n, at)

    def take_range(self, start: int, end: int):
        """
        Take a range of rows from all datasets or collections in this collection.
        This method will fail if :code:`start` < 0, or any of the datasets are not at least
        :code:`end` long.

        Parameters
        ----------
        n: int
            The number of rows to take
        at: str, default = "random"
            The method to use to take rows. Must be one of "start", "end", "random".

        Returns
        -------
        SimulationCollection
            The new simulation collection with only the specified rows.

        """
        if start < 0 or any(len(ds) < end for ds in self.values()):
            raise ValueError(
                "The range must be between zero and the length of the shortest dataset"
            )
        return self.__map("take_range", start, end)

    def with_new_columns(
        self,
        *args,
        datasets: Optional[str | Iterable[str]] = None,
        descriptions: str | dict[str, str] = {},
        **kwargs,
    ):
        """
        Update the datasets within this collection with a set of new columns.
        This method simply calls :py:meth:`opencosmo.Dataset.with_new_columns` or
        :py:meth:`opencosmo.StructureCollection.with_new_columns`, as appropriate.

        You can also optionally pass the "datasets" keyword argument to specify that the
        operation should only be performed on a subset of the datasets.

        If passing in numpy arrays or astropy quantities, they should be provided
        as a dictionary where the keys are the same as the keys in this dataset.

        Parameters
        ----------
        datasets: str | list[str], optional
            The datasets to add the columns to.

        descriptions : str | dict[str, str], optional
            A description for the new columns. These descriptions will be accessible through
            :py:attr:`SimulationCollection(datasets).descriptions <opencosmo.SimulationCollection.descriptions>`.
            If a dictionary, should have keys matching the column names.

        ** columns : opencosmo.DerivedColumn | np.ndarray | units.Quantity
            The new columns
        """
        if datasets is not None:
            if isinstance(datasets, str):
                datasets = [datasets]
            else:
                datasets = list(datasets)

            output = {name: ds for name, ds in self.items()}
            for ds_name in datasets:
                output[ds_name] = output[ds_name].with_new_columns(
                    *args, descriptions=descriptions, **kwargs
                )
            return SimulationCollection(output)

        return self.__map(
            "with_new_columns", *args, descriptions=descriptions, **kwargs
        )

    def evaluate(
        self,
        func: Callable,
        datasets: Optional[str | Iterable[str]] = None,
        format: str = "astropy",
        vectorize: bool = False,
        insert: bool = False,
        **evaluate_kwargs,
    ):
        """
        Evaluate the function :code:`func` on each of the datasets or collections
        held by this SimulationCollection. This function simply delegates to the
        either :py:meth:`StructureCollection.evaluate <opencosmo.StructureCollection.Evaluate>`
        or :py:meth:`Dataset.evaluate <opencosmo.Dataset.Evaluate>` as appropriate. Refer
        to :ref:`Evaluating Complex Expressions on Datasets and Collections` for more details.

        If "datasets" is provided, the evaluation will only be performed on the provided
        datasets.

        Parameters
        ----------

        func: Callable
            The function to evaluate
        datasets: str | list[str], optional
            The datasets to evaluate on. If not provided, will be evaluated on all datasets
        format: str, default = "astropy"
            The format of the data that is provided to your function. If "astropy", will be a dictionary of
            astropy quantities. If "numpy", will be a dictionary of numpy arrays.
        vectorize: bool, default = False
            Whether to vectorize the computation. See :py:meth:`StructureCollection.evaluate <opencosmo.StructureCollection.Evaluate>`
            and/or :py:meth:`Dataset.evaluate <opencosmo.Dataset.Evaluate>` for more details.
        insert: bool, default = True
            Whether or not to insert the results as columns in the datasets. If false, the results will
            be returned directly. If true, this method will return a new Simulation Collection.

        Returns
        -------
        results: SimulationCollection | dict[str, np.ndarray] | dict[str, astropy.units.Quantity]
            The results of the computation, or a new simulation collection with the results inserted.
        """
        if datasets is None:
            datasets = list(self.keys())
        elif isinstance(datasets, str):
            datasets = [datasets]
        else:
            datasets = list(datasets)

        results = self.__map(
            "evaluate",
            func,
            vectorize=vectorize,
            insert=insert,
            format=format,
            construct=insert,
            **evaluate_kwargs,
        )
        if next(iter(results.values())) is None:
            return
        return results

    def sort_by(self, column: str, invert: bool = False):
        """
        Re-order the individual datasets in the collection based on a column. See
        :py:meth:`Dataset.sort_by <opencosmo.Dataset.sort_by>` for usage details.

        Parameters
        ----------
        column : str
            The column in the halo_properties or galaxy_properties dataset to
            order the collection by.

        invert : bool, default = False
            If False (the default) ordering will be done from least to greatest.
            Otherwise greatest to least.

        Returns
        -------
        result : SimulationCollection
            A new SimulationCollection with the datasets ordered by the given column.

        """
        return self.__map("sort_by", column=column, invert=invert)

    def with_units(
        self,
        convention: Optional[str] = None,
        conversions: dict[u.Unit, u.Unit] = {},
        **columns: u.Unit,
    ) -> Self:
        """
        Transform all datasets or collections to use the given unit convention, convert
        all columns with a given unit into a different unit, and/or convert specific column(s)
        to a compatible unit. This method behaves exactly like :meth:`opencosmo.Dataset.with_units`.

        Parameters
        ----------
        convention: str
            The unit convention to use. One of "unitless",
            "scalefree", "comoving", or "physical".

        conversions: dict[astropy.units.Unit, astropy.units.Unit]
            Conversions that apply to all columns in the collection with the
            unit given by the key.

        **column_conversions: astropy.units.Unit
            Custom unit conversions for any column with a specific
            name in the datasets in this collection.

        Returns
        -------
        collection
            A new simulation collection with the requested unit conventions and conversions.


        """
        return self.__map("with_units", convention, conversions=conversions, **columns)



from __future__ import annotations

from collections import defaultdict
from copy import copy
from enum import Enum
from typing import TYPE_CHECKING, Iterable, Mapping, Optional, Type, TypeVar, cast

import h5py
import numpy as np
from mpi4py import MPI

from opencosmo.index import ChunkedIndex
from opencosmo.mpi import get_comm_world

from .schemas import (
    ColumnSchema,
    DatasetSchema,
    EmptyColumnSchema,
    FileSchema,
    LightconeSchema,
    SimCollectionSchema,
    StructCollectionSchema,
    ZeroLengthError,
)

if TYPE_CHECKING:
    from pathlib import Path

    from opencosmo.header import OpenCosmoHeader

    from .protocols import DataSchema

"""
When working with MPI, datasets are chunked across ranks. Here we combine the schemas
from several ranks into a single schema that can be allocated by rank 0. Each 
rank will then write its own data to the specific section of the file 
it is responsible for.

As with schemas and writers, everything is very hierarcical here. A function
does some consistency checks, then calls a function that combines its children.

Ranks with different schemas are supported. For example, one rank may have data for one
dataset in a collection but not another. So long as the top-level structure is the same,
things will be handled.
"""


class CombineState(Enum):
    VALID = 1
    ZERO_LENGTH = 2
    INVALID = 3


def write_parallel(file: Path, file_schema: FileSchema):
    comm = get_comm_world()
    if comm is None:
        raise ValueError("Got a null comm!")
    paths = set(comm.allgather(file))
    if len(paths) != 1:
        raise ValueError("Different ranks recieved a different path to output to!")

    try:
        file_schema.verify()
        results = comm.allgather(CombineState.VALID)
    except ValueError as e:
        results = comm.allgather(CombineState.INVALID)
    except ZeroLengthError:
        results = comm.allgather(CombineState.ZERO_LENGTH)
    if not all(results):
        raise ValueError("One or more ranks recieved invalid schemas!")

    has_data = [i for i, state in enumerate(results) if state == CombineState.VALID]
    group = comm.Get_group()
    new_group = group.Incl(has_data)
    new_comm = comm.Create(new_group)
    if new_comm == MPI.COMM_NULL:
        return cleanup_mpi(comm, new_comm, new_group)
    rank = new_comm.Get_rank()

    new_schema = combine_file_schemas(file_schema, new_comm)
    if rank == 0:
        with h5py.File(file, "w") as f:
            new_schema.allocate(f)

    writer = new_schema.into_writer(new_comm)

    try:
        with h5py.File(file, "a", driver="mpio", comm=new_comm) as f:
            writer.write(f)

    except ValueError:  # parallell hdf5 not available
        raise NotImplementedError(
            "MPI writes without paralell hdf5 are not yet supported"
        )
        nranks = new_comm.Get_size()
        rank = new_comm.Get_rank()
        for i in range(nranks):
            if i == rank:
                with h5py.File(file, "a") as f:
                    writer.write(f)
            new_comm.Barrier()
    cleanup_mpi(comm, new_comm, new_group)


def cleanup_mpi(comm_world: MPI.Comm, comm_write: MPI.Comm, group_write: MPI.Group):
    comm_world.Barrier()
    if comm_write != MPI.COMM_NULL:
        comm_write.Free()
    group_write.Free()


def get_all_child_names(children: dict, comm: MPI.Comm, debug=False):
    child_names = set(children.keys())
    all_child_names: Iterable[str]
    all_child_names = child_names.union(*comm.allgather(child_names))
    all_child_names = list(all_child_names)
    all_child_names.sort()
    return all_child_names


def verify_structure(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    verify_names(schemas, comm)
    verify_types(schemas, comm)


def verify_names(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    names = set(schemas.keys())
    all_names = comm.allgather(names)
    if not all(ns == all_names[0] for ns in all_names[1:]):
        raise ValueError(
            "Tried to combine a collection of schemas with different names!"
        )


def verify_types(schemas: Mapping[str, DataSchema], comm: MPI.Comm):
    types = list(str(type(c)) for c in schemas.values())
    types.sort()
    all_types = comm.allgather(types)
    if not all(ts == all_types[0] for ts in all_types[1:]):
        raise ValueError(
            "Tried to combine a collection of schemas with different types!"
        )


def combine_file_schemas(schema: FileSchema, comm: MPI.Comm) -> FileSchema:
    if comm.Get_size() == 1:
        return schema

    all_child_names = get_all_child_names(
        schema.children if schema is not None else {}, comm
    )
    new_schema = FileSchema()

    for child_name in all_child_names:
        child = schema.children.get(child_name) if schema is not None else None
        new_child = combine_file_child(child, comm)
        new_schema.add_child(new_child, child_name)

    return new_schema


S = TypeVar("S", DatasetSchema, SimCollectionSchema, StructCollectionSchema)


def combine_file_child(schema: S | None, comm: MPI.Comm) -> S:
    match schema:
        case DatasetSchema():
            return cast("S", combine_dataset_schemas(schema, comm))
        case SimCollectionSchema():
            return cast("S", combine_simcollection_schema(schema, comm))
        case StructCollectionSchema():
            return cast("S", combine_structcollection_schema(schema, comm))
        case LightconeSchema():
            return cast("S", combine_lightcone_schema(schema, comm))
        case _:
            raise ValueError(f"Invalid file child of type {type(schema)}")


def validate_headers(
    header: OpenCosmoHeader | None, comm: MPI.Comm, header_updates: dict = {}
):
    all_headers: Iterable[OpenCosmoHeader] = comm.allgather(header)
    all_headers = filter(lambda h: h is not None, all_headers)
    all_headers = list(map(lambda h: h.with_parameters(header_updates), all_headers))
    regions = set([h.file.region for h in all_headers])
    if len(regions) > 1:
        all_headers = [h.with_region(None) for h in all_headers]

    if any(h != all_headers[0] for h in all_headers[1:]):
        raise ValueError("Not all datasets have the same header!")
    return all_headers[0]


def combine_dataset_schemas(
    schema: DatasetSchema | None,
    comm: MPI.Comm,
    header_updates: dict = {},
) -> DatasetSchema:
    if schema is not None:
        header = validate_headers(schema.header, comm, header_updates)
        columns = schema.columns
    else:
        header = validate_headers(None, comm, header_updates)
        columns = defaultdict(dict)

    children = schema.columns if schema is not None else {}
    new_schema = DatasetSchema(header=header)
    all_group_names = get_all_child_names(children, comm)

    for groupname in all_group_names:
        group_column_names = get_all_child_names(children.get(groupname, {}), comm)
        if groupname in ["data", "data_linked"]:
            new_schema.columns[groupname] = combine_data_group(
                columns[groupname], group_column_names, comm
            )
        elif groupname == "index":
            new_schema.columns[groupname] = combine_spatial_index_schema(
                columns[groupname], comm
            )
    return new_schema


def combine_data_group(columns: dict, order: list[str], comm: MPI.Comm):
    output = {}
    for colname in order:
        column = columns.get(colname)
        assert not isinstance(column, EmptyColumnSchema)
        new_column_schema = combine_column_schemas(column, comm)

        output[colname] = new_column_schema
    return output


def combine_spatial_index_schema(
    columns: dict[str, ColumnSchema], comm: MPI.Comm = MPI.COMM_WORLD
):
    has_schema = len(columns) > 0
    all_has_schema = comm.allgather(has_schema)

    if not any(all_has_schema):
        return None

    levels = set(map(lambda key: int(key.split("/")[0][-1]), columns.keys()))
    if not levels:
        n_levels = -1
    else:
        n_levels = max(levels)
    all_max_levels = set(comm.allgather(n_levels))
    if -1 in all_max_levels:
        all_max_levels.remove(-1)

    if len(set(all_max_levels)) != 1:
        raise ValueError("Schemas for all ranks must have the same number of levels!")

    max_level = all_max_levels.pop()
    output = {}
    for level in range(max_level + 1):
        level_schemas = {
            key: val for key, val in columns.items() if f"level_{level}" in key
        }
        new_level_schemas = combine_spatial_index_level_schemas(
            level_schemas, level, comm
        )
        output.update(new_level_schemas)

    return output


def combine_spatial_index_level_schemas(
    schemas: dict[str, ColumnSchema], level: int, comm: MPI.Comm
):
    if schemas:
        assert len(schemas) == 2
        start = schemas[f"level_{level}/start"]
        size = schemas[f"level_{level}/size"]
        start_len = len(start)
        size_len = len(size)
    else:
        start_len = 0
        size_len = 0

    all_start_lens = set(filter(lambda s: s is not None, comm.allgather(start_len)))
    all_size_lens = set(filter(lambda s: s is not None, comm.allgather(size_len)))
    if 0 in all_start_lens:
        all_start_lens.remove(0)
        all_size_lens.remove(0)

    if all_start_lens != all_size_lens or len(all_start_lens) != 1:
        raise ValueError("Invalid starts and sizes")

    level_len = all_start_lens.pop()

    if not schemas:
        source = np.zeros(level_len, dtype=np.int32)
        index = ChunkedIndex.from_size(len(source))
        start = ColumnSchema(
            f"level_{level}/start", index, source, {}, total_length=level_len
        )
        size = ColumnSchema(
            f"level_{level}/size", index, source, {}, total_length=level_len
        )

    else:
        start = ColumnSchema(
            f"level_{level}/start",
            start.index,
            start.source,
            {},
            total_length=level_len,
        )
        size = ColumnSchema(
            f"level_{level}/size",
            size.index,
            size.source,
            {},
            total_length=level_len,
        )

    new_schemas = {f"level_{level}/start": start, f"level_{level}/size": size}
    return new_schemas


def combine_lightcone_schema(schema: LightconeSchema | None, comm: MPI.Comm):
    if schema is None:
        children = {}
    else:
        children = schema.children

    all_child_names = get_all_child_names(children, comm, debug=True)
    new_schema = LightconeSchema()

    for child_name in all_child_names:
        child = children.get(child_name)
        z_range = get_z_range(child, comm)

        new_dataset_schema = combine_dataset_schemas(
            children.get(child_name), comm, {"lightcone/z_range": z_range}
        )
        new_schema.add_child(new_dataset_schema, child_name)
    return new_schema


def get_z_range(ds: DatasetSchema | None, comm: MPI.Comm):
    if ds is not None and ds.header is not None:
        z_ranges = comm.allgather(ds.header.lightcone["z_range"])
    else:
        z_ranges = comm.allgather(None)
    z_ranges = list(filter(lambda dz: dz is not None, z_ranges))
    dzs: Iterable[float] = map(lambda dz: dz[1] - dz[0], z_ranges)
    dzs = list(dzs)
    max_idx = np.argmax(dzs)
    return list(z_ranges)[max_idx]


def combine_simcollection_schema(
    schema: SimCollectionSchema | None, comm: MPI.Comm
) -> SimCollectionSchema:
    if schema is None:
        children = {}
    else:
        children = schema.children
    all_child_names = get_all_child_names(children, comm)

    new_schema = SimCollectionSchema()
    new_child: DatasetSchema | StructCollectionSchema

    for child_name in all_child_names:
        child = children.get(child_name)
        new_child = combine_dataset_schemas(child, comm)
        new_schema.add_child(new_child, child_name)
    return new_schema


def combine_structcollection_schema(
    schema: StructCollectionSchema, comm: MPI.Comm
) -> StructCollectionSchema:
    child_names: Iterable[str] = set(schema.children.keys())
    all_child_names = comm.allgather(child_names)
    if not all(cns == all_child_names[0] for cns in all_child_names[1:]):
        raise ValueError(
            "Tried to combine ismulation collections with different children!"
        )

    child_types = set(str(type(c)) for c in schema.children.values())
    all_child_types = comm.allgather(child_types)
    if not all(cts == all_child_types[0] for cts in all_child_types[1:]):
        raise ValueError(
            "Tried to combine ismulation collections with different children!"
        )

    new_schema = StructCollectionSchema()
    child_names = list(child_names)
    child_names.sort()
    new_child: DatasetSchema | StructCollectionSchema

    for i, name in enumerate(child_names):
        cn = comm.bcast(name)
        child = schema.children[cn]
        if isinstance(child, DatasetSchema):
            new_child = combine_dataset_schemas(child, comm)
        elif isinstance(child, StructCollectionSchema):
            new_child = combine_structcollection_schema(child, comm)
        else:
            raise ValueError(
                "Found a child of a structure collection that was not a Dataset!"
            )
        new_schema.add_child(new_child, cn)

    return new_schema


def verify_column_schemas(schema: ColumnSchema | None, comm: MPI.Comm):
    if schema is None:
        data = comm.allgather(None)
    else:
        data = comm.allgather(
            (schema.source.shape, schema.name, dict(schema.attrs), schema.source.dtype)
        )

    data = list(filter(lambda elem: elem is not None, data))
    if any(d[1:] != data[0][1:] for d in data[1:]):
        raise ValueError("Tried to write incompatible columns to the same output!")
    return data[0]


def combine_column_schemas(
    schema: ColumnSchema | None, comm: MPI.Comm
) -> ColumnSchema | EmptyColumnSchema:
    rank = comm.Get_rank()
    if schema is None:
        length = 0
    else:
        length = len(schema.index)

    shape, name, attrs, dtype = verify_column_schemas(schema, comm)

    lengths = comm.allgather(length)
    total_length = np.sum(lengths)
    rank_offsets = np.insert(np.cumsum(lengths), 0, 0)[:-1]
    rank_offset = rank_offsets[rank]

    new_schema: ColumnSchema | EmptyColumnSchema
    if schema is None:
        new_schema = EmptyColumnSchema(name, attrs, dtype, (total_length,) + shape[1:])
    else:
        new_schema = ColumnSchema(
            schema.name,
            schema.index,
            schema.source,
            schema.attrs,
            total_length=total_length,
        )
        if length != 0:
            new_schema.set_offset(rank_offset)
    return new_schema



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Iterable, NamedTuple, Optional, Protocol, Union

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from opencosmo.index import DataIndex, SimpleIndex
    from opencosmo.spatial.region import BoxRegion
    from opencosmo.units import UnitConvention
    from opencosmo.units.get import UnitApplicator

Point3d = tuple[float, float, float]
Point2d = tuple[float, float]
Points = NDArray[np.number]
SpatialObject = Union["Region", "Points"]


class Region(Protocol):
    """
    The region protocol is intentonally very vague, since we have to
    support both 2d regions and 3d regions.
    """

    def intersects(self, other: "Region") -> bool: ...
    def contains(self, other: SpatialObject): ...
    def into_base_convention(
        self,
        converters: list["UnitApplicator"],
        columns: Iterable[str],
        from_: "UnitConvention",
        unit_kwargs: dict[str, Any],
    ): ...


class Region2d(Region):
    def bounds(self): ...
    def get_healpix_intersections(self, nside: int): ...


class Region3d(Region, Protocol):
    def bounding_box(self) -> "BoxRegion": ...


class TreePartition(NamedTuple):
    idx: DataIndex
    region: Optional[Region]
    level: Optional[int]


class SpatialIndex(Protocol):
    @property
    def subdivision_factor(self) -> int: ...
    def get_partition_region(self, index: SimpleIndex, level: int) -> Region:
        pass

    def query(
        self, region: Region, max_level: int
    ) -> dict[int, tuple[SimpleIndex, SimpleIndex]]:
        """
        Given a region in space, return a dictionary where each key is a level and each
        value is a tuple of DataIndexes. The first DataIndex corresponds to the regions
        that are fully contained by the given region, and the second corresponds to
        regions that only overlap.

        If a given subvolume is full contained by the query region, this method should
        NOT return any sub-sub volumes.
        """
        ...



from typing import ClassVar, Optional

from pydantic import BaseModel, ConfigDict


class DiffskyVersionInfo(BaseModel):
    model_config = ConfigDict(frozen=True)
    ACCESS_PATH: ClassVar[str] = "diffsky_versions"
    diffmah: str
    diffsky: str
    diffstar: str
    diffstarpop: Optional[str] = None
    dsps: str
    jax: str
    numpy: str



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Protocol, Self, Union

if TYPE_CHECKING:
    from opencosmo.dataset import Dataset
    from opencosmo.dataset.column import ColumnMask
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.io.io import OpenTarget
    from opencosmo.io.protocols import DataSchema


class Collection(Protocol):
    """
    Collections represent a group of datasets that are related in some way. They
    support higher-level operations that are applied across all datasets in the
    collection, sometimes in a non-obvious way.

    This protocol defines methods a collection must implement. Most notably they
    must include  __getitem__, keys, values and __items__, which allows
    a collection to behave like a read-only dictionary.


    Note that the "open" and "read" methods are used in the case an entire collection
    is located within a single file. Multi-file collections are handled
    in the collection.io module. Most complexity is hidden from the user
    who simply calls "oc.read" and "oc.open" to get a collection. The io
    module also does sanity checking to ensure files are structurally valid,
    so we do not have to do it here.
    """

    @classmethod
    def open(
        cls, targets: list[OpenTarget], **kwargs
    ) -> Union["Collection", Dataset]: ...

    def make_schema(self) -> DataSchema: ...
    @property
    def dtype(self) -> str | dict[str, str]: ...

    @property
    def header(self) -> OpenCosmoHeader | dict[str, OpenCosmoHeader]: ...

    def __getitem__(self, key: str) -> Union[Dataset, "Collection"]: ...
    def keys(self) -> Iterable[str]: ...
    def values(self) -> Iterable[Union[Dataset, "Collection"]]: ...
    def items(self) -> Iterable[tuple[str, Union[Dataset, "Collection"]]]: ...
    def __enter__(self): ...
    def __exit__(self, *exc_details): ...
    def filter(self, *masks: ColumnMask) -> Self: ...
    def select(self, *args, **kwargs) -> Self: ...
    def with_units(self, convention: str) -> Self: ...
    def take(self, *args, **kwargs) -> Self: ...



from __future__ import annotations

from collections import defaultdict
from copy import copy
from enum import Enum
from functools import cache
from itertools import product
from typing import TYPE_CHECKING, Iterable, Optional, TypeGuard

import numpy as np

import opencosmo as oc
from opencosmo.index import SimpleIndex
from opencosmo.spatial.region import BoxRegion

if TYPE_CHECKING:
    from opencosmo.spatial.protocols import Region
    from opencosmo.spatial.region import Point3d

Index3d = tuple[int, int, int]


"""
In an oct tree, the space is subdivided into octants. At level one, the space is 
subdivided into 8 octants with indexes (0, 0, 0) -> (1, 1, 1). At the next level, we 
have 64 octants labeled (0,0,0) -> (4,4,4) and so on.

To query, we traverse recursively. If the octant is completely enclosed by the query 
region, we simply return a version of that octant with no children. If the octant 
itersects the query region, we call the function on the octant's children. We then 
return a copy of an octant WITH the children that 

To evaluate the tree, we again traverse it recursively. If an octant has no children, 
we know all objects in that octant should be included in the output. Otherwise, we move 
on to the children.

However at the lowest level of the octant this breaks down. Here we instead get all the 
data for all of the octants, and check if they are contained by our query region.

"""


@cache
def get_index3d(p: Point3d, level: int, box_size: float) -> Index3d:
    block_size = box_size / (2**level)
    return int(p[0] // block_size), int(p[1] // block_size), int(p[2] // block_size)


def get_octtree_index(idx: Index3d, level: int) -> int:
    oct_idx = 0
    idx_ = copy(idx)
    for i in range(level):
        oct_idx |= (idx_[0] & 1) << 3 * i
        oct_idx |= (idx_[1] & 1) << (3 * i + 1)
        oct_idx |= (idx_[2] & 1) << (3 * i + 2)
        idx_ = (idx_[0] >> 1, idx_[1] >> 1, idx_[2] >> 1)
    return oct_idx


def get_3d_index(oct_idx: int, level: int) -> tuple:
    """
    Inverts the z-order curve indexing to retrieve the 3D index (i, j, k) from the
    1D index.

    Args:
        oct_idx (int): The 1D index obtained from the z-order curve.
        level (int): The depth level of the OctTree.

    Returns:
        tuple: The 3D index (i, j, k).
    """
    i, j, k = 0, 0, 0  # Initialize the 3D indices
    for le in range(level):
        # Extract the bits for i, j, and k at the current level
        i |= ((oct_idx >> (3 * le)) & 1) << le
        j |= ((oct_idx >> (3 * le + 1)) & 1) << le
        k |= ((oct_idx >> (3 * le + 2)) & 1) << le
    return i, j, k


def get_octant(oct_idx: int, level: int, box_size: float) -> Octant:
    index3d = get_3d_index(oct_idx, level)
    nside = 2**level
    octant_size = box_size / nside
    center = tuple(octant_size * (i + 0.5) for i in index3d)
    return Octant(index3d, center, octant_size / 2)


def get_children(idx: Index3d) -> Iterable[Index3d]:
    return (
        (idx[0] * 2 + dx, idx[1] * 2 + dy, idx[2] * 2 + dz)
        for dx, dy, dz in product(range(2), repeat=3)
    )


def get_region(octants: list[Octant]) -> BoxRegion:
    """
    Return the region that encloses this list of octants
    """
    if len(octants) == 1:
        return octants[0].bounding_box()
    bounds = map(lambda oct: np.array(oct.bounding_box().bounds), octants)
    bound_arr = np.stack(list(bounds))

    mins = bound_arr[:, :, 0].min(axis=0)
    maxs = bound_arr[:, :, 1].max(axis=0)
    return oc.make_box(tuple(mins), tuple(maxs))


class OctTreeIndex:
    subdivision_factor = 8

    def __init__(self, root: Octant):
        """
        An octree index is used to spatialy index snapshot data.
        """
        self.root = root

    def get_partition_region(self, index: SimpleIndex, level: int):
        octants = [
            get_octant(idx, level, 2 * self.root.halfwidth)
            for idx in index.into_array()
        ]
        return get_region(octants)

    @classmethod
    def from_box_size(cls, box_size: int):
        halfwidth = box_size / 2
        root = Octant((0, 0, 0), (halfwidth, halfwidth, halfwidth), halfwidth)
        return OctTreeIndex(root)

    def query(
        self, region: Region, max_level: int
    ) -> dict[int, tuple[SimpleIndex, SimpleIndex]]:
        containment: dict[Octant, Intersection] = {}
        new_root = self.root.query(region, 0, max_level, containment)
        if new_root is None:
            raise ValueError("Query region is not within the actual simulation box!")
        output = {}
        for level, (cidx, iidx) in make_octree_indices(
            new_root, 0, containment
        ).items():
            output[level] = (SimpleIndex(cidx), SimpleIndex(iidx))

        return output


def make_octree_indices(
    oct: Octant, current_level: int, containment: dict[Octant, Intersection]
) -> dict[int, tuple[np.ndarray, np.ndarray]]:
    """
    Returns two arrays per level. One for regions that are fully contained.
    The other for regions that only intersect.
    """
    oct_intersection = containment[oct]
    current_level_output = (np.array([], dtype=int), np.array([], dtype=int))
    idx = get_octtree_index(oct.idx, current_level)
    output: dict[int, tuple[np.ndarray, np.ndarray]] = defaultdict(
        lambda: (np.array([], dtype=int), np.array([], dtype=int))
    )
    if oct_intersection == Intersection.CONTAINED:
        current_level_output = (np.array([idx], dtype=int), np.array([], dtype=int))
        output[current_level] = current_level_output

    elif oct_intersection == Intersection.INTERSECTS and len(oct.children) == 0:
        current_level_output = (np.array([], dtype=int), np.array([idx], dtype=int))
        output[current_level] = current_level_output

    elif oct_intersection == Intersection.INTERSECTS:
        for child in oct.children:
            child_output = make_octree_indices(child, current_level + 1, containment)
            for key, indices in child_output.items():
                current_output = output[key]
                new_contains = np.append(current_output[0], indices[0])
                new_intersects = np.append(current_output[1], indices[1])
                output[key] = (new_contains, new_intersects)

    return output


class Intersection(Enum):
    NONE = 0
    INTERSECTS = 1
    CONTAINED = 2


class Octant:
    def __init__(
        self,
        idx: Index3d,
        center: Point3d,
        halfwidth: float,
        children: Optional[list[Octant]] = None,
    ):
        self.idx = idx
        self.center = center
        self.halfwidth = halfwidth
        self.children = children if children is not None else []

    def __repr__(self):
        return f"{self.idx}: {self.bounding_box()}"

    def __hash__(self):
        return hash((self.idx, self.center, self.halfwidth))

    def __eq__(self, other):
        if not isinstance(other, Octant):
            return False

        output = (
            self.idx == other.idx
            and self.center == other.center
            and self.halfwidth == other.halfwidth
        )
        return output

    def make_children(self):
        if len(self.children) == 8:
            return

        child_halfwidth = self.halfwidth / 2.0
        for z, y, x in product(range(2), repeat=3):
            child_idx = (self.idx[0] * 2 + x, self.idx[1] * 2 + y, self.idx[2] * 2 + z)
            child_center = (
                self.center[0] + child_halfwidth * (2 * x - 1),
                self.center[1] + child_halfwidth * (2 * y - 1),
                self.center[2] + child_halfwidth * (2 * z - 1),
            )
            child = Octant(child_idx, child_center, child_halfwidth)
            self.children.append(child)

    @cache
    def bounding_box(self):
        p1 = tuple(ci - self.halfwidth for ci in self.center)
        p2 = tuple(ci + self.halfwidth for ci in self.center)

        return oc.make_box(p1, p2)

    def query(
        self,
        region: Region,
        current_level: int,
        max_level: int,
        containment: dict[Octant, Intersection],
    ) -> Optional[Octant]:
        if not isinstance(region, BoxRegion):
            raise ValueError("Did not recieve a 3D region!")
        if region.contains(self.bounding_box()):
            containment[self] = Intersection.CONTAINED
            return Octant(self.idx, self.center, self.halfwidth)
        if region.intersects(self.bounding_box()):
            containment[self] = Intersection.INTERSECTS
            if current_level == max_level:
                return Octant(self.idx, self.center, self.halfwidth)

            self.make_children()
            queried_children = map(
                lambda reg: reg.query(
                    region, current_level + 1, max_level, containment
                ),
                self.children,
            )

            def is_not_none(reg: Octant | None) -> TypeGuard[Octant]:
                return reg is not None

            new_children: list[Octant] = list(filter(is_not_none, queried_children))
            return Octant(self.idx, self.center, self.halfwidth, new_children)

        containment[self] = Intersection.NONE
        return None



from .collection import (
    Lightcone,
    SimulationCollection,
    StructureCollection,
    open_linked_files,
)
from .dataset import Dataset, col
from .io import open, write
from .spatial import make_box, make_cone

__all__ = [
    "write",
    "col",
    "open",
    "Dataset",
    "StructureCollection",
    "SimulationCollection",
    "Lightcone",
    "open_linked_files",
    "make_box",
    "make_cone",
]



import inspect
from typing import Callable, ClassVar, Type

from astropy import cosmology
from pydantic import BaseModel, Field, computed_field


def make_cosmology(parameters: "CosmologyParameters") -> cosmology.Cosmology:
    cosmology_type = get_cosmology_type(parameters)
    expected_arguments = inspect.signature(cosmology_type).parameters.keys()
    input_paremeters = {}
    for argname in expected_arguments:
        try:
            input_paremeters[argname] = getattr(parameters, argname)
        except AttributeError:
            continue
    return cosmology_type(**input_paremeters)


def get_cosmology_type(parameters: "CosmologyParameters") -> Type[cosmology.Cosmology]:
    is_flat = (parameters.Om0 + parameters.Ode0) == 1.0
    if parameters.w0 == -1 and parameters.wa == 0:
        if is_flat:
            return cosmology.FlatLambdaCDM
        else:
            return cosmology.LambdaCDM
    if parameters.w0 != -1 and parameters.wa == 0:
        if is_flat:
            return cosmology.FlatwCDM
        else:
            return cosmology.wCDM
    if parameters.wa != 0:
        if is_flat:
            return cosmology.Flatw0waCDM
        else:
            return cosmology.w0waCDM

    raise ValueError("Could not determine cosmology type.")


class CosmologyParameters(BaseModel, frozen=True):
    """
    Responsible for validating cosmology parameters and handling differences in
    naming conventions between OpenCosmo and astropy.cosmology. Generally should
    not be used by the user directly
    """

    ACCESS_PATH: ClassVar[str] = "cosmology"
    ACCESS_TRANSFORMATION: ClassVar[Callable] = make_cosmology

    h: float = Field(ge=0.0, description="Reduced Hubble constant")

    @computed_field  # type: ignore
    @property
    def H0(self) -> float:
        """
        Hubble constant in km/s/Mpc
        """
        return self.h * 100

    Om0: float = Field(ge=0.0, description="Total matter density", alias="omega_m")
    Ob0: float = Field(ge=0.0, description="Baryon density", alias="omega_b")
    Ode0: float = Field(ge=0.0, description="Dark energy density", alias="omega_l")
    Neff: float = Field(
        gt=0.0, description="Effective number of neutrinos", alias="n_eff_massless"
    )
    n_eff_massive: float = Field(
        0, ge=0.0, description="Effective number of massive neutrinos"
    )
    sigma_8: float = Field(ge=0.0, description="RMS mass fluctuation at 8 Mpc/h")
    w0: float = Field(description="Dark energy equation of state", alias="w_0")
    wa: float = Field(
        description="Dark energy equation of state evolution", alias="w_a"
    )



from __future__ import annotations

import os
from logging import getLogger
from typing import TYPE_CHECKING, Optional

import networkx as nx

from .source import install_conda_forge, install_github, install_pip
from .specs import get_specs

if TYPE_CHECKING:
    from .specs import DependencySpec

CONDA_ENV = os.environ.get("CONDA_DEFAULT_ENV")
logger = getLogger("opencosmo")


def install_spec(name: str, versions: dict[str, Optional[str]] = {}, dev: bool = False):
    spec = get_specs()[name]
    logger.info(f"Installing analysis package {name}")
    requirements = spec.requirements
    raw_graph = {name: dep.depends_on for name, dep in requirements.items()}
    graph = nx.DiGraph(raw_graph).reverse()
    transaction = {}
    method: Optional[str] = None
    dev_transaction: dict[str, str | None] = {}

    for requirement, data in requirements.items():
        if requirement not in versions and not data.optional:
            versions[requirement] = data.version

    for node in nx.topological_sort(graph):
        if node not in versions:
            continue
        version = versions.get(node)
        if version is not None and "dev" in version:
            dev_transaction[node] = version
            continue

        requirement_method = resolve_method(
            versions.get(node), requirements[node].prefer_source
        )
        if method is None or requirement_method == method:
            method = requirement_method
            transaction.update({node: versions.get(node)})
            continue
        execute_transaction(method, transaction, requirements, dev)
        method = requirement_method
        transaction = {node: versions.get(node)}
    execute_transaction(method, transaction, requirements, dev)
    execute_transaction("pip-git", dev_transaction, requirements, dev)


def resolve_method(version: Optional[str], method: Optional[str]):
    if version is None or "+g" not in version:
        if method in (None, "conda-forge") and CONDA_ENV is not None:
            return "conda-forge"
    elif "+g" in version:
        return "pip-git"
    elif method is not None:
        return method
    return "pip"


def execute_transaction(
    method: Optional[str],
    transaction: dict[str, Optional[str]],
    requirements: dict[str, DependencySpec],
    dev: bool = False,
):
    print(f"Installing {','.join(transaction.keys())}")
    if method == "conda-forge":
        install_conda_forge(transaction)
    elif method == "pip-git":
        for name, version in transaction.items():
            assert version is not None
            install_github(name, version, requirements)
    elif method == "pip":
        install_pip(transaction, dev)



from __future__ import annotations

from typing import TYPE_CHECKING, TypeGuard

import h5py
import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray


def get_data_simple(data: h5py.Dataset | np.ndarray, index: NDArray[np.int_]):
    if isinstance(data, np.ndarray):
        return data[index]

    min = index.min()
    max = index.max()
    remaining_shape = data.shape[1:]
    length = max + 1 - min

    shape = (length,) + remaining_shape

    buffer = np.zeros(shape, data.dtype)

    data.read_direct(buffer, np.s_[min : max + 1], np.s_[0:length])
    return buffer[index - min]


def get_data_chunked(
    data: h5py.Dataset | np.ndarray, starts: NDArray[np.int_], sizes: NDArray[np.int_]
):
    """
    We assume that starts are ordered, and chunks are non-overlapping

    """

    shape = (np.sum(sizes),) + data.shape[1:]
    storage = np.zeros(shape, dtype=data.dtype)
    running_index = 0
    for i, (start, size) in enumerate(zip(starts, sizes)):
        source_slice = np.s_[start : start + size]
        dest_slice = np.s_[running_index : running_index + size]

        if isinstance(data, h5py.Dataset):
            data.read_direct(storage, source_slice, dest_slice)
        else:
            storage[dest_slice] = data[source_slice]

        running_index += size
    return storage



from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Callable, Iterable, Sequence

from astropy.table import Column, QTable  # type: ignore

from opencosmo.evaluate import insert, make_output_from_first_values, prepare_kwargs

if TYPE_CHECKING:
    from opencosmo import Dataset

"""
Although the user-facing name for this operation is "evaluate", the pattern 
we are using here is known as a "visitor."
"""


def visit_dataset(
    function: Callable,
    dataset: "Dataset",
    vectorize: bool = False,
    format: str = "astropy",
    evaluator_kwargs: dict[str, Any] = {},
):
    __verify(function, dataset, evaluator_kwargs.keys())
    dataset = __prepare(function, dataset, evaluator_kwargs.keys())
    if vectorize:
        result = __visit_vectorize(function, dataset, format, evaluator_kwargs)
        if result is not None and not isinstance(result, dict):
            return {function.__name__: result}
        return result
    else:
        kwargs, iterable_kwargs = prepare_kwargs(len(dataset), evaluator_kwargs)
        return __visit_rows(function, dataset, format, kwargs, iterable_kwargs)


def __visit_rows(
    function: Callable,
    dataset: "Dataset",
    format="astropy",
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, Sequence] = {},
):
    requested_columns = (
        set(signature(function).parameters)
        - set(kwargs.keys())
        - set(iterable_kwargs.keys())
    )
    using_all_columns = len(dataset.columns) > 1 and len(requested_columns) == 1

    first_row_kwargs = kwargs | {name: arr[0] for name, arr in iterable_kwargs.items()}
    storage = __make_output(function, dataset, using_all_columns, first_row_kwargs)
    for i, row in enumerate(dataset.rows(output=format)):
        if i == 0:
            continue
        iter_kwargs = {name: arr[i] for name, arr in iterable_kwargs.items()}
        if using_all_columns:
            output = function(row, **kwargs, **iter_kwargs)
        else:
            output = function(**row, **kwargs, **iter_kwargs)
        if storage is not None:
            insert(storage, i, output)
    return storage


def __make_output(
    function: Callable,
    dataset: "Dataset",
    using_all_columns: bool,
    kwargs: dict[str, Any],
) -> dict | None:
    if using_all_columns:
        first_values = function(next(dataset.take(1, at="start").rows()), **kwargs)
    else:
        first_values = function(**next(dataset.take(1, at="start").rows()), **kwargs)
    n_rows = len(dataset)
    if first_values is None:
        return None
    if not isinstance(first_values, dict):
        name = function.__name__
        first_values = {name: first_values}

    return make_output_from_first_values(first_values, n_rows)


def __visit_vectorize(
    function: Callable,
    dataset: "Dataset",
    format: str = "astropy",
    evaluator_kwargs: dict[str, Any] = {},
):
    data = dataset.get_data(format)
    if format == "astropy" and isinstance(data, QTable):
        data = {name: col for name, col in data.items()}
    elif isinstance(data, Column):
        data = {data.name: data.quantity}

    if not isinstance(data, dict) or (
        len(data) > 1 and len(signature(function).parameters) == 1
    ):
        return function(data, **evaluator_kwargs)
    return function(**data, **evaluator_kwargs)


def __prepare(function: Callable, dataset: "Dataset", evaluator_kwargs: Iterable[str]):
    function_arguments = set(signature(function).parameters.keys())

    input_columns = function_arguments.intersection(dataset.columns)
    if len(input_columns) == 0 and len(function_arguments) == 1:
        return dataset
    return dataset.select(input_columns)


def __verify(function: Callable, dataset: "Dataset", kwarg_names: Iterable[str]):
    function_signature = signature(function)
    required_parameters = set()
    for name, parameter in function_signature.parameters.items():
        if parameter.default == Parameter.empty:
            required_parameters.add(name)

    dataset_columns = set(dataset.columns)
    kwarg_names = set(kwarg_names)
    missing = required_parameters - dataset_columns - kwarg_names
    if not missing:
        return
    elif len(missing) > 1:
        raise ValueError(
            f"All inputs to the function must either be column names or passed as keyword arguments! Found unknown input(s) {','.join(missing)}"
        )



# ruff: noqa: TC001 TC003
from enum import Enum
from typing import TYPE_CHECKING, Optional

from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)

from opencosmo.spatial import models as sm
from opencosmo.units import UnitConvention


def empty_string_to_none(value: str) -> Optional[str]:
    if type(value) is str and value == "":
        return None
    return value


class FileType(Enum):
    galaxy_properties = "galaxy_properties"
    galaxy_particles = "galaxy_particles"
    halo_properties = "halo_properties"
    halo_profiles = "halo_profiles"
    halo_particles = "halo_particles"
    diffsky_fits = "diffsky_fits"


class FileParameters(BaseModel):
    model_config = ConfigDict(use_enum_values=True, frozen=True)
    origin: str = "HACC"
    data_type: FileType
    is_lightcone: bool
    redshift: float
    step: int
    region: Optional[sm.RegionModel] = None
    unit_convention: UnitConvention = UnitConvention.SCALEFREE

    @model_validator(mode="before")
    @classmethod
    def empty_string_to_none(cls, data):
        if isinstance(data, dict):
            data = {k: empty_string_to_none(v) for k, v in data.items()}
        return data

    @model_validator(mode="before")
    def parse_region(cls, data):
        region_keys = list(filter(lambda k: k.startswith("region_"), data.keys()))
        region_dict = {k[7:]: data.pop(k) for k in region_keys}
        if region_dict:
            data.update({"region": region_dict})

        return data

    @field_validator("is_lightcone", mode="before")
    def validate_is_lightcone(cls, value):
        return bool(value)

    @field_validator("unit_convention", mode="before")
    def validate_convention(cls, value):
        if isinstance(value, str):
            return UnitConvention(value)
        return value

    @field_serializer("unit_convention")
    def serialize_convention(self, value):
        if isinstance(value, UnitConvention):
            return value.value
        return value

    @model_serializer(mode="wrap")
    def serialize_model(self, handle):
        dump = handle(self)
        if dump["region"] is not None:
            region = dump.pop("region")
            region = {f"region_{k}": v for k, v in region.items()}
            dump.update(region)
        return dump



from __future__ import annotations

from functools import cached_property, reduce
from itertools import chain
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable, Optional, Self

import astropy.units as u  # type: ignore
import numpy as np
from astropy.table import Column, vstack  # type: ignore

import opencosmo as oc
from opencosmo.dataset.column import DerivedColumn
from opencosmo.evaluate import prepare_kwargs
from opencosmo.index import SimpleIndex
from opencosmo.io.io import open_single_dataset
from opencosmo.io.schemas import LightconeSchema

if TYPE_CHECKING:
    from astropy.coordinates import SkyCoord
    from astropy.cosmology import Cosmology
    from astropy.table import Table

    from opencosmo.dataset import Dataset
    from opencosmo.dataset.column import ColumnMask
    from opencosmo.header import OpenCosmoHeader
    from opencosmo.io.io import OpenTarget
    from opencosmo.parameters.hacc import HaccSimulationParameters
    from opencosmo.spatial import Region


def get_redshift_range(datasets: list[Dataset]):
    redshift_ranges = [ds.header.lightcone["z_range"] for ds in datasets]
    if all(rr is not None for rr in redshift_ranges):
        min_redshift = min(rr[0] for rr in redshift_ranges)
        max_redshift = max(rr[1] for rr in redshift_ranges)

    else:
        steps = np.fromiter((ds.header.file.step for ds in datasets), dtype=int)
        step_zs = datasets[0].header.simulation["step_zs"]
        min_step = np.min(steps)
        max_step = np.max(steps)

        min_redshift = step_zs[max_step]
        max_redshift = step_zs[min_step - 1]
    return (min_redshift, max_redshift)


def is_in_range(dataset: Dataset, z_low: float, z_high: float):
    z_range = dataset.header.lightcone["z_range"]
    if z_range is None:
        z_range = get_redshift_range([dataset])
    if z_high < z_range[0] or z_low > z_range[1]:
        return False
    return True


def sort_table(table: Table, column: str, invert: bool):
    column_data = table[column]
    if invert:
        column_data = -column_data
    indices = np.argsort(column_data)
    for name in table.columns:
        table[name] = table[name][indices]
    return table


def take_from_sorted(
    lightcone: "Lightcone", sort_by: str, invert: bool, n: int, at: str | int
):
    column = np.concatenate(
        [ds.select(sort_by).get_data("numpy") for ds in lightcone.values()]
    )
    if invert:
        column = -column
    sort_index = np.argsort(column)
    if at == "start":
        sort_index = sort_index[:n]
    elif at == "end":
        sort_index = sort_index[-n:]
    elif isinstance(at, int):
        if at + n > len(sort_index) or at < 0:
            raise ValueError(
                "Requested a range that is outside the size of this dataset!"
            )
        sort_index = sort_index[at : at + n]

    sorted_indices = np.sort(sort_index)
    return sorted_indices


def with_redshift_column(dataset: Dataset):
    """
    Ensures a column exists called "redshift" which contains the redshift of the objects
    in the lightcone.
    """
    if "redshift" in dataset.columns:
        return dataset

    elif "fof_halo_center_a" in dataset.columns:
        z_col = 1 / oc.col("fof_halo_center_a") - 1
        return dataset.with_new_columns(redshift=z_col)
    elif "redshift_true" in dataset.columns:
        z_col = 1 * oc.col("redshift_true")
        return dataset.with_new_columns(redshift=z_col)
    raise ValueError(
        "Unable to find a redshift or scale factor column for this lightcone dataset"
    )


class Lightcone(dict):
    """
    A lightcone contains two or more datasets that are part of a lightcone. Typically
    each dataset will cover a specific redshift range. The Lightcone object
    hides these details, providing an API that is identical to the standard
    Dataset API. Additionally, the lightcone contains some convinience functions
    for standard operations.
    """

    def __init__(
        self,
        datasets: dict[str, Dataset],
        z_range: Optional[tuple[float, float]] = None,
        hidden: Optional[set[str]] = None,
        ordered_by: Optional[tuple[str, bool]] = None,
    ):
        datasets = {k: with_redshift_column(ds) for k, ds in datasets.items()}
        self.update(datasets)
        z_range = (
            z_range
            if z_range is not None
            else get_redshift_range(list(datasets.values()))
        )

        columns: set[str] = reduce(
            lambda left, right: left.union(set(right.columns)), self.values(), set()
        )
        if len(columns) != len(next(iter(self.values())).columns):
            raise ValueError("Not all lightcone datasets have the same columns!")
        header = next(iter(self.values())).header
        self.__header = header.with_parameter("lightcone/z_range", z_range)
        if hidden is None:
            hidden = set()

        self.__hidden = hidden
        self.__ordered_by = ordered_by

    def __repr__(self):
        """
        A basic string representation of the dataset
        """
        length = len(self)

        if len(self) < 10:
            repr_ds = self
            table_head = ""
        else:
            repr_ds = self.take(10, at="start")
            table_head = "First 10 rows:\n"

        table_repr = repr_ds.data.__repr__()
        # remove the first line
        table_repr = table_repr[table_repr.find("\n") + 1 :]
        z_range = self.z_range
        head = (
            f"OpenCosmo Lightcone Dataset (length={length}, "
            f"{z_range[0]} < z < {z_range[1]})\n"
        )
        cosmo_repr = f"Cosmology: {self.cosmology.__repr__()}" + "\n"
        return head + cosmo_repr + table_head + table_repr

    def __len__(self):
        return sum(len(ds) for ds in self.values())

    def __enter__(self):
        return self

    def __exit__(self, *exc_details):
        for dataset in self.values():
            try:
                dataset.close()
            except ValueError:
                continue

    @property
    def header(self) -> OpenCosmoHeader:
        """
        The header associated with this dataset.

        OpenCosmo headers generally contain information about the original data this
        dataset was produced from, as well as any analysis that was done along
        the way.

        Returns
        -------
        header: opencosmo.header.OpenCosmoHeader

        """
        return self.__header

    @property
    def columns(self) -> list[str]:
        """
        The names of the columns in this dataset.

        Returns
        -------
        columns: list[str]
        """
        cols = next(iter(self.values())).columns
        cols = list(filter(lambda col: col not in self.__hidden, cols))
        return cols

    @cached_property
    def descriptions(self) -> dict[str, Optional[str]]:
        """
        Return the descriptions (if any) of the columns in this lightcone as a dictonary.
        Columns without a description will be included in the dictionary with a value
        of None

        Returns
        -------

        descriptions : dict[str, str | None]
            The column descriptions
        """
        descriptions = next(iter(self.values())).descriptions
        descriptions = dict(
            filter(lambda kv: kv[0] not in self.__hidden, descriptions.items())
        )
        return descriptions

    @property
    def cosmology(self) -> Cosmology:
        """
        The cosmology of the simulation this dataset is drawn from as
        an astropy.cosmology.Cosmology object.

        Returns
        -------
        cosmology: astropy.cosmology.Cosmology
        """
        return self.__header.cosmology

    @property
    def dtype(self) -> str:
        """
        The data type of this dataset.

        Returns
        -------
        dtype: str
        """
        return self.__header.file.data_type

    @property
    def region(self) -> Region:
        """
        The region this dataset is contained in. If no spatial
        queries have been performed, this will be the entire
        simulation box for snapshots or the full sky for lightcones

        Returns
        -------
        region: opencosmo.spatial.Region

        """
        return next(iter(self.values())).region

    @property
    def simulation(self) -> HaccSimulationParameters:
        """
        The parameters of the simulation this dataset is drawn
        from.

        Returns
        -------
        parameters: opencosmo.parameters.hacc.HaccSimulationParameters
        """
        return self.__header.simulation

    @property
    def z_range(self):
        """
        The redshift range of this lightcone.

        Returns
        -------
        z_range: tuple[float, float]
        """

        return self.__header.lightcone["z_range"]

    def get_data(self, output="astropy"):
        """
        Get the data in this dataset as an astropy table/column or as
        numpy array(s). Note that a dataset does not load data from disk into
        memory until this function is called. As a result, you should not call
        this function until you have performed any transformations you plan to
        on the data.

        You can get the data in two formats, "astropy" (the default) and "numpy".
        "astropy" format will return the data as an astropy table with associated
        units. "numpy" will return the data as a dictionary of numpy arrays. The
        numpy values will be in the associated unit convention, but no actual
        units will be attached.

        If the dataset only contains a single column, it will be returned as an
        astropy.table.Column or a single numpy array.

        This method does not cache data. Calling "get_data" always reads data
        from disk, even if you have already called "get_data" in the past.
        You can use :py:attr:`Dataset.data <opencosmo.Lightcone.data>` to return
        data and keep it in memory.

        Parameters
        ----------
        output: str, default="astropy"
            The format to output the data in

        Returns
        -------
        data: Table | Column | dict[str, ndarray] | ndarray
            The data in this dataset.
        """
        if output not in {"astropy", "numpy"}:
            raise ValueError(f"Unknown output type {output}")

        data = [ds.get_data(unpack=False) for ds in self.values()]
        table = vstack(data, join_type="exact")
        if self.__ordered_by is not None:
            table.sort(self.__ordered_by[0], reverse=self.__ordered_by[1])

        table.remove_columns(self.__hidden)

        if len(table.colnames) == 1:
            table = next(table.itercols())

        if output == "numpy":
            if isinstance(table, (u.Quantity, Column)):
                return table.value
            else:
                return {name: col.value for name, col in table.items()}

        return table

    @property
    def data(self):
        """
        Return the data in the dataset in astropy format. The value of this
        attribute is equivalent to the return value of
        :code:`Dataset.get_data("astropy")`. However data retrieved via this
        attribute will be cached, meaning further calls to
        :py:attr:`Dataset.data <opencosmo.Lightcone.data>` should be instantaneous.

        However there is one caveat. If you modify the table, those modifications will
        persist if you later request the data again with this attribute. Calls to
        :py:meth:`Lightcone.get_data <opencosmo.Lightcone.get_data>` will be unaffected,
        and datasets generated from this dataset will not contain the modifications.
        If you plan to modify the data in this table, you should use
        :py:meth:`Lightcone.with_new_columns <opencosmo.Lightcone.with_new_columns>`.


        Returns
        -------
        data : astropy.table.Table or astropy.table.Column
            The data in the dataset.

        """
        return self.get_data("astropy")

    @classmethod
    def open(cls, targets: list[OpenTarget], **kwargs):
        datasets: dict[str, Dataset] = {}

        for target in targets:
            ds = open_single_dataset(target)
            if not isinstance(ds, Lightcone) or len(ds.keys()) != 1:
                raise ValueError(
                    "Lightcones can only contain datasets (not collections)"
                )
            if target.group.name != "/":
                key = target.group.name.split("/")[-1]
            else:
                key = f"{target.header.file.step}_{target.header.file.data_type}"
            datasets[key] = next(iter(ds.values()))

        return cls(datasets)

    def with_redshift_range(self, z_low: float, z_high: float):
        """
        Restrict this lightcone to a specific redshift range. Lightcone datasets will
        always contain a column titled "redshift." This function is always operates on
        this column.

        This function also updates the value in
        :py:meth:`Lightcone.z_range <opencosmo.collection.Lightcone.z_range>`,
        so you should always use it rather than filteringo n the column directly.
        """
        z_range = self.__header.lightcone["z_range"]
        if z_high < z_low:
            z_high, z_low = z_low, z_high

        if z_high < z_range[0] or z_low > z_range[1]:
            raise ValueError(
                f"This lightcone only ranges from z = {z_range[0]} to z = {z_range[1]}"
            )

        elif z_low == z_high:
            raise ValueError("Low and high values of the redshift range are the same!")
        new_datasets = {}
        for key, dataset in self.items():
            if not is_in_range(dataset, z_low, z_high):
                continue
            new_dataset = dataset.filter(
                oc.col("redshift") > z_low, oc.col("redshift") < z_high
            )
            if len(new_dataset) > 0:
                new_datasets[key] = new_dataset
        return Lightcone(
            new_datasets, (z_low, z_high), self.__hidden, self.__ordered_by
        )

    def __map(
        self,
        method,
        *args,
        hidden: Optional[set[str]] = None,
        mapped_arguments: dict[str, dict[str, Any]] = {},
        construct: bool = True,
        **kwargs,
    ):
        """
        This type of collection will only ever be constructed if all the underlying
        datasets have the same data type, so it is always safe to map operations
        across all of them.
        """
        output = {}
        hidden = hidden if hidden is not None else self.__hidden
        for ds_name, dataset in self.items():
            dataset_mapped_arguments = {
                arg_name: args[ds_name] for arg_name, args in mapped_arguments.items()
            }
            output[ds_name] = getattr(dataset, method)(
                *args, **kwargs, **dataset_mapped_arguments
            )

        if construct:
            return Lightcone(output, self.z_range, hidden, self.__ordered_by)
        return output

    def __map_attribute(self, attribute):
        return {k: getattr(v, attribute) for k, v in self.items()}

    def make_schema(self) -> LightconeSchema:
        schema = LightconeSchema()
        for name, dataset in self.items():
            ds_schema = dataset.make_schema()
            ds_schema.header = ds_schema.header.with_parameter(
                "lightcone/z_range", self.z_range
            )
            schema.add_child(ds_schema, name)
        return schema

    def bound(self, region: Region, select_by: Optional[str] = None):
        """
        Restrict the dataset to some subregion. The subregion will always be evaluated
        in the same units as the current dataset. For example, if the dataset is
        in the default "comoving" unit convention, positions are always in units of
        comoving Mpc. However Region objects themselves do not carry units.
        See :doc:`spatial_ref` for details of how to construct regions.

        Parameters
        ----------
        region: opencosmo.spatial.Region
            The region to query.

        Returns
        -------
        dataset: opencosmo.Dataset
            The portion of the dataset inside the selected region

        Raises
        ------
        ValueError
            If the query region does not overlap with the region this dataset resides
            in
        AttributeError:
            If the dataset does not contain a spatial index
        """
        return self.__map("bound", region, select_by)

    def cone_search(self, center: tuple | SkyCoord, radius: float | u.Quantity):
        """
        Perform a search for objects within some angular distance of some
        given point on the sky. This is a convinience function around
        :py:meth:`bound <opencosmo.Lightcone.bound>` and is exactly
        equivalent to

        .. code-block:: python

            region = oc.make_cone(center, radius)
            ds = ds.bound(region)

        Parameters
        ----------
        center: tuple | SkyCoord
            The center of the region to search. If a tuple and no units are provided
            assumed to be RA and Dec in degrees.

        radius: float | astropy.units.Quantity
            The angular radius of the region to query. If no units are provided,
            assumed to be degrees.

        Returns
        -------
        new_lightcone: opencosmo.Lightcone
            The rows in this lightcone that fall within the given region.

        """
        region = oc.make_cone(center, radius)
        return self.bound(region)

    def evaluate(
        self,
        func: Callable,
        format: str = "astropy",
        vectorize=False,
        insert=True,
        **evaluate_kwargs,
    ):
        """
        Iterate over the rows in this collection, apply `func` to each, and collect
        the result as new columns in the dataset. You may also choose to simply return thevalues
        instead of inserting them as a column

        This function is the equivalent of :py:meth:`with_new_columns <opencosmo.Lightcone.with_new_columns>`
        for cases where the new column is not a simple algebraic combination of existing columns. Unlike
        :code:`with_new_columns`, this method will evaluate the results immediately and the resulting
        columns will not change under unit transformations.

        The function should take in arguments with the same name as the columns in this dataset that
        are needed for the computation, and should return a dictionary of output values.
        The dataset will automatically selected the needed columns to avoid reading unnecessarily reading
        data from disk. The new columns will have the same names as the keys of the output dictionary
        See :ref:`Evaluating On Datasets` for more details.

        If vectorize is set to True, the full columns will be pased to the dataset. Otherwise,
        rows will be passed to the function one at a time.

        This function behaves identically to :py:meth:`Dataset.evaluate <opencosmo.Dataset.evaluate>`

        Parameters
        ----------
        func: Callable
            The function to evaluate on the rows in the dataset.

        format: str, default = "astropy"
            The format of the data that is provided to your function. If "astropy", will be a dictionary of
            astropy quantities. If "numpy", will be a dictionary of numpy arrays.

        vectorize: bool, default = False
            Whether to provide the values as full columns (True) or one row at a time (False)

        insert: bool, default = True
            If true, the data will be inserted as a column in this dataset. Otherwise the data will be returned.

        Returns
        -------
        dataset : Lightcone
            The new lightcone dataset with the evaluated column(s)
        """
        kwargs, iterable_kwargs = prepare_kwargs(len(self), evaluate_kwargs)
        iterable_kwargs_by_dataset = {}
        indices = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))[
            :-1
        ]
        for name, arr in iterable_kwargs.items():
            splits = np.array_split(arr, indices)
            iterable_kwargs_by_dataset[name] = dict(zip(self.keys(), splits))

        result = self.__map(
            "evaluate",
            func=func,
            format=format,
            vectorize=vectorize,
            insert=insert,
            mapped_arguments=iterable_kwargs_by_dataset,
            construct=insert,
            **kwargs,
        )
        if next(iter(result.values())) is None:
            return

        if insert:
            assert isinstance(result, Lightcone)
            return result

        keys = next(iter(result.values())).keys()
        output = {}
        for key in keys:
            output[key] = np.concatenate([r[key] for r in result.values()])
        return output

    def filter(self, *masks: ColumnMask, **kwargs) -> Self:
        """
        Filter the dataset based on some criteria. See :ref:`Querying Based on Column
        Values` for more information.

        Parameters
        ----------
        *masks : Mask
            The masks to apply to dataset, constructed with :func:`opencosmo.col`

        Returns
        -------
        dataset : Dataset
            The new dataset with the masks applied.

        Raises
        ------
        ValueError
            If the given  refers to columns that are
            not in the dataset, or the  would return zero rows.

        """
        return self.__map("filter", *masks, **kwargs)

    def rows(self) -> Generator[dict[str, float | u.Quantity], None, None]:
        """
        Iterate over the rows in the dataset. Rows are returned as a dictionary
        For performance, it is recommended to first select the columns you need to
        work with.

        Yields
        -------
        row : dict
            A dictionary of values for each row in the dataset with units.
        """
        yield from chain.from_iterable(v.rows() for v in self.values())

    def select(self, columns: str | Iterable[str]) -> Self:
        """
        Create a new dataset from a subset of columns in this dataset.

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to select.

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected columns.

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """
        if isinstance(columns, str):
            columns = [columns]
        columns = set(columns)
        hidden = self.__hidden

        if "redshift" not in columns:
            columns.add("redshift")
            hidden = hidden.union({"redshift"})

        if self.__ordered_by is not None and self.__ordered_by[0] not in columns:
            columns.add(self.__ordered_by[0])
            hidden = hidden.union({self.__ordered_by[0]})

        return self.__map("select", columns, hidden=hidden)

    def drop(self, columns: str | Iterable[str]) -> Self:
        """
        Produce a new dataset by dropping columns from this dataset.

        Parameters
        ----------
        columns : str or list[str]
            The column or columns to drop.

        Returns
        -------
        dataset : Dataset
            The new dataset without the dropped columns

        Raises
        ------
        ValueError
            If any of the given columns are not in the dataset.
        """
        if isinstance(columns, str):
            columns = [columns]

        dropped_columns = set(columns)
        current_columns = set(self.columns)
        if missing := dropped_columns.difference(current_columns):
            raise ValueError(
                f"Tried to drop columns that are not in this dataset: {missing}"
            )
        kept_columns = current_columns - dropped_columns
        return self.select(kept_columns)

    def take(self, n: int, at: str = "random") -> "Lightcone":
        """
        Create a new dataset from some number of rows from this dataset.

        Can take the first n rows, the last n rows, or n random rows
        depending on the value of 'at'.

        Parameters
        ----------
        n : int
            The number of rows to take.
        at : str
            Where to take the rows from. One of "start", "end", or "random".
            The default is "random".

        Returns
        -------
        dataset : Dataset
            The new dataset with only the selected rows.

        Raises
        ------
        ValueError
            If n is negative or greater than the number of rows in the dataset,
            or if 'at' is invalid.

        """
        if n > len(self):
            raise ValueError(
                "Number of rows to take must be less than number of rows in dataset"
            )
        if at == "random":
            rs = 0
            indices = np.random.choice(len(self), n, replace=False)
            indices = np.sort(indices)
            return self.__take_rows(indices)

        elif self.__ordered_by is not None:
            index = take_from_sorted(self, *self.__ordered_by, n=n, at=at)
            return self.__take_rows(index)
        elif at == "start":
            return self.take_range(0, n)
        elif at == "end":
            return self.take_range(len(self) - n, len(self))
        else:
            raise ValueError(
                f'"at" should be one of ("start", "end", "random", got {at}'
            )

    def take_range(self, start: int, end: int):
        """
        Create a new lightcone from a row range in this lightcone. We use standard
        indexing conventions, so the rows included will be start -> end - 1. Because
        lightcones are stacked by redshift, this operation effectively takes a
        redshift range. If you know the exact redshift range you want, use
        :py:meth:`with_redshift_range <opencosmo.Lightcone.with_redshift_range>`.

        Parameters
        ----------
        start : int
            The beginning of the range
        end : int
            The end of the range

        Returns
        -------
        lightcone : opencosmo.Lightcone
            The lightcone with only the specified range of rows.

        Raises
        ------
        ValueError
            If start or end are negative or greater than the length of the dataset
            or if end is greater than start.

        """
        if start < 0 or end > len(self):
            raise ValueError("Got row indices that are out of range!")

        if self.__ordered_by is not None:
            indices = take_from_sorted(self, *self.__ordered_by, end - start, at=start)
            return self.__take_rows(indices)

        ends = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))
        starts = np.insert(ends, 0, 0)[:-1]
        clipped_starts = np.clip(starts, a_min=start, a_max=None)
        clipped_ends = np.clip(ends, a_min=None, a_max=end)

        output = {}
        for i, (name, dataset) in enumerate(self.items()):
            if starts[i] == clipped_starts[i] and ends[i] == clipped_ends[i]:
                output[name] = dataset
            elif clipped_starts[i] >= clipped_ends[i]:
                continue
            else:
                output[name] = dataset.take_range(
                    clipped_starts[i] - starts[i], clipped_ends[i] - starts[i]
                )
        return Lightcone(output, self.z_range, self.__hidden, self.__ordered_by)

    def take_rows(self, rows: np.ndarray):
        """
        Take the rows of a lightcone specified by the :code:`rows` argument.
        :code:`rows` should be an array of integers.

        Parameters
        ----------
        rows : np.ndarray[int]
            The indices of the rows to take.

        Returns
        -------
        dataset: The dataset with only the specified rows included

        Raises:
        -------
        ValueError:
            If any of the indices is less than 0 or greater than the length of the
            lightcone.

        """
        rows = np.sort(rows)
        if rows[-1] >= len(self) or rows[0] < 0:
            raise ValueError(
                "Rows must be between 0 and the length of this dataset - 1"
            )
        if self.__ordered_by is not None:
            data = np.concatenate(
                [
                    ds.select(self.__ordered_by[0]).get_data("numpy")
                    for ds in self.values()
                ]
            )
            if self.__ordered_by[1]:
                data = -data
            sort_index = np.argsort(data)
            rows = sort_index[rows]
            rows.sort()

        return self.__take_rows(rows)

    def __take_rows(self, rows: np.ndarray):
        """
        Takes rows from this lightcone while ignoring sort. "rows" is assumed to be sorte.
        For internal use only.
        """
        ds_ends = np.cumsum(np.fromiter((len(ds) for ds in self.values()), dtype=int))
        partitions = np.searchsorted(rows, ds_ends)
        splits = np.split(rows, partitions)
        output = {**self}
        rs = 0
        for split, (name, dataset) in zip(splits, self.items()):
            if len(split) > 0:
                output[name] = dataset.take_rows(split - rs)
            rs += len(dataset)
        return Lightcone(output, self.z_range, self.__hidden, self.__ordered_by)

    def with_new_columns(
        self,
        descriptions: str | dict[str, str] = {},
        **columns: DerivedColumn | np.ndarray | u.Quantity,
    ):
        """
        Create a new dataset with additional columns. These new columns can be derived
        from columns already in the dataset, a numpy array, or an Astropy quantity
        array. When a column is derived from other columns, it will behave
        appropriately under unit transformations. See :ref:`Adding Custom Columns`
        and :py:meth:`Dataset.with_new_columns <opencosmo.Dataset.with_new_columns>`
        for examples.

        Parameters
        ----------
        descriptions : str | dict[str, str], optional
            A description for the new columns. These descriptions will be accessible through
            :py:attr:`Lightcone.descriptions <opencosmo.Lighcone.descriptions>`. If a dictionary,
            should have keys matching the column names.

        ** columns : opencosmo.DerivedColumn | np.ndarray | u.quantity
            The new columns

        Returns
        -------
        dataset : opencosmo.Dataset
            This dataset with the columns added

        """
        derived = {}
        raw = {}
        for name, column in columns.items():
            if isinstance(column, DerivedColumn):
                derived[name] = column
            elif len(column) != len(self):
                raise ValueError(
                    f"New column {name} has length {len(column)} but this dataset "
                    f"has length {len(self)}"
                )
            else:
                raw[name] = column

        split_points = np.cumsum([len(ds) for ds in self.values()])
        split_points = np.insert(0, 0, split_points)[:-1]
        raw_split = {name: np.split(arr, split_points) for name, arr in raw.items()}
        new_datasets = {}
        for i, (ds_name, ds) in enumerate(self.items()):
            raw_columns = {name: arrs[i] for name, arrs in raw_split.items()}
            columns_input = raw_columns | derived
            new_dataset = ds.with_new_columns(descriptions, **columns_input)
            new_datasets[ds_name] = new_dataset
        return Lightcone(new_datasets, self.z_range, self.__hidden, self.__ordered_by)

    def sort_by(self, column: str, invert: bool = False):
        """
        Sort this dataset by the values in a given column. By default sorting is in
        ascending order (least to greatest). Pass invert = True to sort in descending
        order (greatest to least).

        This can be used to, for example, select largest halos in a given
        dataset:

        .. code-block:: python

            dataset = oc.open("haloproperties.hdf5")
            dataset = dataset
                        .sort_by("fof_halo_mass")
                        .take(100, at="start")

        Parameters
        ----------
        column : str
            The column in the halo_properties or galaxy_properties dataset to
            order the collection by.

        invert : bool, default = False
            If False (the default), ordering will be from least to greatest.
            Otherwise greatest to least.

        Returns
        -------
        result : Dataset
            A new Dataset ordered by the given column.


        """

        if column not in self.columns:
            raise ValueError(f"Column {column} does not exist in this dataset!")
        return Lightcone(dict(self), self.z_range, self.__hidden, (column, invert))

    def with_units(
        self,
        convention: Optional[str] = None,
        conversions: dict[u.Unit, u.Unit] = {},
        **columns: u.Unit,
    ) -> Self:
        r"""
        Create a new lightcone from this one with a different unit convention or
        with certain columns converted to a different compatible unit.

        Unit conversions are always performed after a change of convention, and
        changing conventions clears any existing unit conversions.

        For more, see :doc:`units`.

        .. code-block:: python

            import astropy.units as u

            # this works
            lc = lc.with_units(fof_halo_mass=u.kg)

            # this clears the previous conversion
            lc = lc.with_units("scalefree")

            # This now fails, because the units of masses
            # are Msun / h, which cannot be converted to kg
            lc = lc.with_units(fof_halo_mass=u.kg)

            # this will now work, wince the units of halo mass in the "physical"
            # convention are Msun (no h).
            lc = lc.with_units("physical", fof_halo_mass=u.kg, fof_halo_center_x=u.lyr)

            # Suppose you want your distances in lightyears, but the x coordinate of your
            # halo center in kilometers, for some reason ¯\_(ツ)_/¯
            blanket_conversions = {u.Mpc: u.lyr}
            lc = lc.with_units(conversions = blanket_conversions, fof_halo_center_x = u.km)

        Parameters
        ----------
        convention : str, optional
            The unit convention to use. One of "physical", "comoving",
            "scalefree", or "unitless".

        conversions: dict[astropy.units.Unit, astropy.units.Unit]
            Conversions that apply to all columns in the lightcone with the
            unit given by the key.

        **column_conversions: astropy.units.Unit
            Custom unit conversions for specific columns
            in this dataset.

        Returns
        -------
        lightcone : Lightcone
            The new lightcone with the requested unit convention and/or conversions.
        """
        return self.__map(
            "with_units",
            convention=convention,
            conversions=conversions,
            **columns,
        )



from __future__ import annotations

from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Callable, Iterable, Mapping, Optional, Sequence

import numpy as np
from astropy.units import Quantity  # type: ignore

from opencosmo import dataset as ds
from opencosmo.dataset.evaluate import visit_dataset
from opencosmo.evaluate import insert, make_output_from_first_values, prepare_kwargs

if TYPE_CHECKING:
    from numpy.typing import DTypeLike

    from opencosmo import StructureCollection


def visit_structure_collection(
    function: Callable,
    spec: Mapping[str, Optional[list[str]]],
    collection: StructureCollection,
    format: str = "astropy",
    dtype: Optional[DTypeLike] = None,
    evaluator_kwargs: dict[str, Any] = {},
):
    spec = dict(spec)
    __verify(function, spec, collection, evaluator_kwargs.keys())
    to_visit = __prepare_collection(spec, collection)
    kwargs, iterable_kwargs = prepare_kwargs(len(collection), evaluator_kwargs)
    if dtype is None:
        dtype = np.float64

    storage = __make_output(function, to_visit, format, kwargs, iterable_kwargs)
    for i, structure in enumerate(to_visit.objects()):
        if i == 0:
            continue
        iterable_kwarg_values = {name: arr[i] for name, arr in iterable_kwargs.items()}
        input_structure = __make_input(structure, format)

        output = function(**input_structure, **kwargs, **iterable_kwarg_values)
        if storage is not None:
            insert(storage, i, output)

    return storage


def __make_input(structure: dict, format: str = "astropy"):
    values = {}
    for name, element in structure.items():
        if isinstance(element, dict):
            values[name] = __make_input(element, format)
        elif isinstance(element, ds.Dataset):
            data = element.get_data(format)
            values[name] = data
        elif isinstance(element, Quantity) and format == "numpy":
            values[name] = element.value
        else:
            values[name] = element
    return values


def __make_output(
    function: Callable,
    collection: StructureCollection,
    format: str = "astropy",
    kwargs: dict[str, Any] = {},
    iterable_kwargs: dict[str, Sequence] = {},
) -> dict | None:
    first_structure = next(collection.take(1, at="start").objects())
    first_input = __make_input(first_structure, format)
    first_values = function(
        **first_input,
        **kwargs,
        **{name: arr[0] for name, arr in iterable_kwargs.items()},
    )
    if first_values is None:
        return None
    if not isinstance(first_values, dict):
        name = function.__name__
        first_values = {name: first_values}
    n_rows = len(collection)
    return make_output_from_first_values(first_values, n_rows)


def __prepare_collection(
    spec: dict[str, Optional[list[str]]], collection: StructureCollection
) -> StructureCollection:
    collection = collection.with_datasets(list(spec.keys()))
    selections = {ds_name: cols for ds_name, cols in spec.items() if cols is not None}
    collection = collection.select(**selections)
    return collection


def __verify(
    function: Callable,
    spec: dict[str, Optional[list[str]]],
    collection: StructureCollection,
    kwarg_keys: Iterable[str],
):
    datasets_in_collection = set(collection.keys())
    kwarg_keys = set(kwarg_keys)
    fn_signature = signature(function)
    parameters = fn_signature.parameters

    for name, param in parameters.items():
        if name not in spec and name not in kwarg_keys and param == Parameter.empty:
            spec.update({name: None})

    datasets_in_spec = set(spec.keys())

    if not datasets_in_spec.issubset(datasets_in_collection):
        raise ValueError(
            "This collection is missing datasets "
            f"{datasets_in_spec - datasets_in_collection} requested for this visitor"
        )
    for ds_name, columns_in_spec in spec.items():
        if columns_in_spec is None:
            continue
        dataset = collection[ds_name]
        if not isinstance(dataset, ds.Dataset):
            if not isinstance(columns_in_spec, dict):
                raise ValueError(
                    "When passing columns to a nested structure collection, the argument should be a dictionary"
                )
                for key, value in columns_in_spec.items():
                    if key not in dataset.keys():
                        raise ValueError(
                            "No dataset {key} found in structure collection"
                        )
                    elif set(dataset[key].columns).difference(value):
                        raise ValueError(
                            "Missing some requested columns in this datset!"
                        )
            continue

        columns_to_check = set(columns_in_spec)
        columns_in_dataset = set(dataset.columns)
        if not columns_to_check.issubset(columns_in_dataset):
            raise ValueError(
                "Dataset {ds_name} is missing columns "
                f"{columns_to_check - columns_in_dataset} requested for this visitor"
            )

    if not datasets_in_spec.issubset(parameters.keys()):
        raise ValueError(
            "Visitor function must use the names of the datasets it requests as its "
            "argument names"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Tuple

import numpy as np
import yt  # type: ignore
from matplotlib.colors import LogNorm  # type: ignore
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar  # type: ignore
from unyt import unyt_quantity  # type: ignore
from yt.visualization.base_plot_types import get_multi_plot  # type: ignore

import opencosmo as oc
from opencosmo.analysis import create_yt_dataset

if TYPE_CHECKING:
    from matplotlib.figure import Figure
    from yt.visualization.plot_window import NormalPlot, PlotWindow

# ruff: noqa: E501


def ParticleProjectionPlot(
    *args, **kwargs
) -> yt.AxisAlignedProjectionPlot | yt.OffAxisProjectionPlot:
    """
    Wrapper for `yt.ParticleProjectionPlot <https://yt-project.org/doc/reference/api/yt.visualization.plot_window.html#yt.visualization.plot_window.ParticleProjectionPlot>`_.

    Creates a 2D projection plot of particle-based data along a specified axis.

    Parameters
    ----------
    *args :
        Positional arguments passed directly to `yt.ParticleProjectionPlot`.
    **kwargs :
        Keyword arguments passed directly to `yt.ParticleProjectionPlot`.

    Returns
    -------
    yt.visualization.plot_window.ParticleProjectionPlot
        A ParticleProjectionPlot object containing the particle projection plot.
    """
    # mypy gets this wrong. ParticleProjectionPlot is basically a factory class
    return yt.ParticleProjectionPlot(*args, **kwargs)  # type: ignore


def ProjectionPlot(*args, **kwargs) -> NormalPlot:
    """
    Wrapper for `yt.ProjectionPlot <https://yt-project.org/doc/reference/api/yt.visualization.plot_window.html#yt.visualization.plot_window.ProjectionPlot>`_.

    Creates a 2D projection plot of particle-based data along a specified axis.
    Smoothing is applied to SPH particle data over the smoothing length

    Parameters
    ----------
    *args :
        Positional arguments passed directly to `yt.ProjectionPlot`.
    **kwargs :
        Keyword arguments passed directly to `yt.ProjectionPlot`.

    Returns
    -------
    yt.visualization.plot_window.ProjectionPlot
        A ProjectionPlot object containing the smoothed particle projection plot.
    """
    return yt.ProjectionPlot(*args, **kwargs)


def SlicePlot(*args, **kwargs) -> NormalPlot:
    """
    Wrapper for `yt.SlicePlot <https://yt-project.org/doc/reference/api/yt.visualization.plot_window.html#yt.visualization.plot_window.SlicePlot>`_.

    Creates a 2D slice plot of particle-based data along a specified axis.
    Smoothing is applied to SPH particle data over the smoothing length

    Parameters
    ----------
    *args :
        Positional arguments passed directly to `yt.SlicePlot`.
    **kwargs :
        Keyword arguments passed directly to `yt.SlicePlot`.

    Returns
    -------
    yt.visualization.plot_window.PlotWindow
        A PlotWindow object containing the particle slice plot.
    """
    return yt.SlicePlot(*args, **kwargs)


def ProfilePlot(*args, **kwargs) -> yt.ProfilePlot:
    """
    Wrapper for `yt.ProfilePlot <https://yt-project.org/doc/reference/api/yt.visualization.particle_plots.html#yt.visualization.particle_plots.ParticleProjectionPlot>`_.

    Creates a bin-averaged profile of a dependent variable
    as a function of one or more independent variables.

    Parameters
    ----------
    *args :
        Positional arguments passed directly to `yt.ProfilePlot`.
    **kwargs :
        Keyword arguments passed directly to `yt.ProfilePlot`.

    Returns
    -------
    yt.visualization.plot_window.PlotWindow
        A PlotWindow object containing the profile plot.
    """
    return yt.ProfilePlot(*args, **kwargs)


def PhasePlot(*args, **kwargs) -> yt.PhasePlot:
    """
    Wrapper for `yt.PhasePlot <https://yt-project.org/doc/reference/api/yt.visualization.profile_plotter.html#yt.visualization.profile_plotter.PhasePlot>`_.

    Creates a 2D histogram (phase plot) showing how one quantity varies as a function
    of two others, useful for visualizing thermodynamic or structural relationships
    (e.g., temperature vs. density colored by mass).

    Parameters
    ----------
    *args :
        Positional arguments passed directly to `yt.PhasePlot`.
    **kwargs :
        Keyword arguments passed directly to `yt.PhasePlot`.

    Returns
    -------
    yt.visualization.plot_window.PlotWindow
        A PlotWindow object containing the phase plot.
    """
    return yt.PhasePlot(*args, **kwargs)


def visualize_halo(
    halo_id: int,
    data: oc.StructureCollection,
    projection_axis: Optional[str] = "z",
    length_scale: Optional[str] = "top left",
    text_color: Optional[str] = "gray",
    width: Optional[float] = None,
) -> Figure:
    """
    Creates a figure showing particle projections of dark matter, stars, gas, and/or gas temperature
    for given halo. If any of the listed particle types are not present in the dataset, this will
    create a horizontal arrangement with only the particles/fields that are present. Otherwise,
    creates a 2x2-panel figure. Each panel is an 800x800 pixel array.

    To customize the arrangement of panels, fields, colormaps, etc., see
    :func:`halo_projection_array`.


    Parameters
    ----------
    halo_id : int
        Identifier of the halo to be visualized.
    data : opencosmo.StructureCollection
        OpenCosmo StructureCollection object containing both halo properties and particle data
        (e.g. output of ``opencosmo.open([haloproperties, sodbighaloparticles])``).
    projection_axis : str, optional
        Data is projected along this axis (``"x"``, ``"y"``, or ``"z"``).
        Overridden if ``params["projection_axes"]`` is provided
    length_scale : str or None, optional
        Optionally add a horizontal bar denoting length scale in Mpc.

        Options:
            - ``"top left"``: add to top left panel
            - ``"top right"``: add to top right panel
            - ``"bottom left"``: add to bottom left panel
            - ``"bottom right"``: add to bottom right panel
            - ``"all top"``: add to all panels on top row
            - ``"all bottom"``: add to all panels on bottom row
            - ``"all left"``: add to all panels on leftmost column
            - ``"all right"``: add to all panels on rightmost column
            - ``"all"``: add to all panels
            - ``None``: no length scale on any panel

    text_color : str, optional
        Set the color of all text annotations. Default is "gray"
    width : float, optional
        Width of each projection panel in units of R200 for the halo.
        If None, plots full subvolume around halo.

    Returns
    -------
    matplotlib.figure.Figure
        A matplotlib Figure object.
    """

    params: Dict[str, Any] = {
        "fields": [],
        "weight_fields": [],
        "zlims": [],
        "labels": [],
        "cmaps": [],
    }

    ptypes = [
        key.removesuffix("_particles")
        for key in data.keys()
        if key.endswith("_particles")
    ]

    any_supported = False

    if "dm" in ptypes:
        any_supported = True
        params["fields"].append(("dm", "particle_mass"))
        params["weight_fields"].append(None)
        params["zlims"].append(None)
        params["labels"].append("Dark Matter")
        params["cmaps"].append("gray")
    elif "gravity" in ptypes:
        any_supported = True
        # particle mass not stored for GO simulations because each particle has the same mass.
        # Use particle_ones for making images in this case instead
        params["fields"].append(("gravity", "particle_ones"))
        params["weight_fields"].append(None)
        params["zlims"].append(None)
        params["labels"].append("Dark Matter")
        params["cmaps"].append("gray")

    if "star" in ptypes:
        any_supported = True
        params["fields"].append(("star", "particle_mass"))
        params["weight_fields"].append(None)
        params["zlims"].append(None)
        params["labels"].append("Stars")
        params["cmaps"].append("bone")

    if "gas" in ptypes:
        any_supported = True
        params["fields"].append(("gas", "particle_mass"))
        params["weight_fields"].append(None)
        params["zlims"].append(None)
        params["labels"].append("Gas")
        params["cmaps"].append("viridis")
        # temperature field should always exist if gas
        # particles are present
        params["fields"].append(("gas", "temperature"))
        params["weight_fields"].append(("gas", "density"))
        params["zlims"].append((1e7, 1e8))
        params["labels"].append("Gas Temperature")
        params["cmaps"].append("inferno")

    if not any_supported:
        raise RuntimeError(
            "No compatible particle types present in dataset for this function. "
            'Possible options are "dm", "gravity", "star", and "gas".'
        )

    halo_ids: list[int] | tuple[list[int], list[int]]
    if len(params["fields"]) == 4:
        # if 4 fields, make a 2x2 figure
        halo_ids = ([halo_id, halo_id], [halo_id, halo_id])
        params = {key: (value[:2], value[2:]) for key, value in params.items()}

    else:
        # otherwise, do 1xN
        halo_ids = np.shape(params["fields"])[0] * [halo_id]
        params = {key: [value] for key, value in params.items()}

    return halo_projection_array(
        halo_ids,
        data,
        params=params,
        length_scale=length_scale,
        width=width,
        projection_axis=projection_axis,
        text_color=text_color,
    )


def halo_projection_array(
    halo_ids: int | list[int] | tuple[list[int], list[int]] | np.ndarray,
    data: oc.StructureCollection,
    field: Optional[Tuple[str, str]] = ("dm", "particle_mass"),
    weight_field: Optional[Tuple[str, str]] = None,
    projection_axis: Optional[str] = "z",
    cmap: Optional[str] = "gray",
    zlim: Optional[Tuple[float, float]] = None,
    params: Optional[Dict[str, Any]] = None,
    length_scale: Optional[str] = None,
    text_color: Optional[str] = "gray",
    width: Optional[float] = None,
) -> Figure:
    """
    Creates a multipanel figure of projections for different fields and/or halos.

    By default, creates an arrangement of dark matter particle projections with the
    same shape as `halo_ids`. Each panel is an 800x800 pixel array.

    Customizable — can change which fields are plotted for which halos, their order,
    weighting, etc., using `params`.

    **NOTE:** Dark matter particle masses often aren't stored for gravity-only simulations
    because the particles all have the same mass by construction. The particles are also
    labelled as "gravity" particles in this case instead of "dm" particles in the data.
    To project dark matter particles in gravity only, one can use the ``("gravity", "particle_ones")``
    field in place of ``("dm", "particle_mass")``. This will produce the same final image.

    Parameters
    ----------
    halo_ids : int or 2D array of int
        Unique ID of the halo(s) to be visualized. The shape of `halo_ids` sets the layout
        of the figure (e.g., if `halo_ids` is a 2x3 array, the outputted figure will be a 2x3
        array of projections). If `int`, a single panel is output while preserving formatting.
    data : opencosmo.StructureCollection
        OpenCosmo StructureCollection dataset containing both halo properties and particle data
        (e.g., output of ``opencosmo.open([haloproperties, sodbighaloparticles])``).
    field : tuple of str, optional
        Field to plot for all panels. Follows yt naming conventions (e.g., ``("dm", "particle_mass")``,
        ``("gas", "temperature")``). Overridden if ``params["fields"]`` is provided.
    weight_field : tuple of str, optional
        Field to weight by during projection. Follows yt naming conventions.
        Overridden if ``params["weight_fields"]`` is provided.
    projection_axis : str, optional
        Data is projected along this axis (``"x"``, ``"y"``, or ``"z"``).
        Overridden if ``params["projection_axes"]`` is provided
    cmap : str
        Matplotlib colormap to use for all panels. Overridden if ``params["cmaps"]`` is provided.
        See https://matplotlib.org/stable/gallery/color/colormap_reference.html for named colormaps.
    zlim : tuple of float, optional
        Colorbar limits for `field`. Overridden if ``params["zlims"]`` is provided.
    length_scale : str or None, optional
        Optionally add a horizontal bar denoting length scale in Mpc.

        Options:
            - ``"top left"``: add to top left panel
            - ``"top right"``: add to top right panel
            - ``"bottom left"``: add to bottom left panel
            - ``"bottom right"``: add to bottom right panel
            - ``"all top"``: add to all panels on top row
            - ``"all bottom"``: add to all panels on bottom row
            - ``"all left"``: add to all panels on leftmost column
            - ``"all right"``: add to all panels on rightmost column
            - ``"all"``: add to all panels
            - ``None``: no length scale shown

    params : dict, optional
        Dictionary of customization parameters for the projection panels. Overrides
        defaults. All values must be 2D arrays with the same shape as `halo_ids`.

        Keys may include:
            - ``"fields"``: 2D array of fields to plot (yt naming conventions)
            - ``"weight_fields"``: 2D array of projection weights (or None)
            - ``"projection_axes"``: 2D array of projection axes ("x", "y", or "z")
            - ``"zlims"``: 2D array of colorbar limits (log-scaled)
            - ``"labels"``: 2D array of panel labels (or None)
            - ``"cmaps"``: 2D array of Matplotlib colormaps for each panel
            - ``"widths"``: 2D array of widths in units of R200
    text_color : str, optional
        Set the color of all text annotations. Default is "gray"
    width : float, optional
        Width of each projection panel in units of R200 for the halo.
        Overridden if ``params["widths"]`` is provided.
        If None, plots full subvolume.

    Returns
    -------
    matplotlib.figure.Figure
        A Matplotlib Figure object.
    """

    halo_ids = np.atleast_2d(halo_ids)

    # determine shape of figure
    fig_shape = np.shape(halo_ids)

    # Default plotting parameters
    if weight_field is None:
        weight_field_ = np.full(fig_shape, None)
    else:
        weight_field_ = np.reshape(
            [weight_field for _ in range(np.prod(fig_shape))],
            (fig_shape[0], fig_shape[1], 2),
        )

    if zlim is None:
        zlim_ = np.full(fig_shape, None)
    else:
        zlim_ = np.reshape(
            [zlim for _ in range(np.prod(fig_shape))], (fig_shape[0], fig_shape[1], 2)
        )

    default_params = {
        "fields": (
            np.reshape(
                [field for _ in range(np.prod(fig_shape))],
                (fig_shape[0], fig_shape[1], 2),
            )
        ),
        "weight_fields": (weight_field_),
        "zlims": (zlim_),
        "projection_axes": (np.full(fig_shape, projection_axis)),
        "labels": (np.full(fig_shape, None)),
        "cmaps": (np.full(fig_shape, cmap)),
        "widths": (np.full(fig_shape, width)),
    }

    # Override defaults with user-supplied params (if any)
    params = params or {}

    fields = params.get("fields", default_params["fields"])
    weight_fields = params.get("weight_fields", default_params["weight_fields"])
    projection_axes = params.get("projection_axes", default_params["projection_axes"])
    zlims = params.get("zlims", default_params["zlims"])
    labels = params.get("labels", default_params["labels"])
    cmaps = params.get("cmaps", default_params["cmaps"])
    widths = params.get("widths", default_params["widths"])

    nrow, ncol = fig_shape

    ilen, jlen = None, None

    # define figure and axes
    fig, axes, cbars = get_multi_plot(fig_shape[1], fig_shape[0], cbar_padding=0)

    # are we plotting a single halo multiple times?
    halo_ids = np.array(halo_ids)
    halo_id_previous = np.inf

    for i in range(nrow):
        for j in range(ncol):
            halo_id = halo_ids[i][j]

            # retrieve halo particle info if new halo
            if (i == 0 and j == 0) or halo_id != halo_id_previous:
                # retrieve properties of halo
                data_id = data.filter(oc.col("unique_tag") == halo_id)
                halo_data = next(iter(data_id.objects()))

                # load particles into yt
                ds = create_yt_dataset(halo_data)

            halo_properties = halo_data["halo_properties"]

            Rh = unyt_quantity.from_astropy(halo_properties["sod_halo_radius"])

            field, weight_field, zlim, width = (
                tuple(fields[i][j]),
                weight_fields[i][j],
                zlims[i][j],
                widths[i][j],
            )

            if weight_field is not None:
                weight_field = tuple(weight_field)  # type: ignore
            if zlim is not None:
                zlim = tuple(zlim)  # type: ignore

            label = labels[i][j]

            proj = ParticleProjectionPlot(
                ds, projection_axes[i][j], field, weight_field=weight_field
            )

            proj.set_background_color(field, color="black")

            if width is not None:
                width_ = width
                proj.set_width(width_ * Rh)
            else:
                width_ = (max(ds.domain_width.to("Mpc")) / Rh).d  # type: ignore

            # fetch figure buffer (2D array of pixel values)
            # and re-plot on each panel with imshow
            frb = proj.frb

            ax = axes[i][j]

            if zlim is not None:
                zmin, zmax = zlim
            else:
                zmin, zmax = None, None

            ax.imshow(
                frb[field],
                origin="lower",
                cmap=cmaps[i][j],
                norm=LogNorm(vmin=zmin, vmax=zmax),
            )
            ax.set_facecolor("black")

            ax.xaxis.set_visible(False)
            ax.yaxis.set_visible(False)
            cbars[i].set_visible(False)

            if label is not None:
                # add panel label
                ax.text(
                    0.06,
                    0.94,
                    label,
                    transform=ax.transAxes,
                    ha="left",
                    va="top",
                    fontsize=12,
                    fontfamily="DejaVu Serif",
                    color=text_color,
                )

            if length_scale is not None:
                match length_scale:
                    case "top left":
                        ilen, jlen = 0, 0
                    case "top right":
                        ilen, jlen = 0, ncol - 1
                    case "bottom left":
                        ilen, jlen = nrow - 1, 0
                    case "bottom right":
                        ilen, jlen = nrow - 1, ncol - 1
                    case "all left":
                        ilen, jlen = i, 0
                    case "all right":
                        ilen, jlen = i, ncol - 1
                    case "all top":
                        ilen, jlen = 0, j
                    case "all bottom":
                        ilen, jlen = nrow - 1, j
                    case "all":
                        ilen, jlen = i, j

                if i == ilen and j == jlen:
                    # add length scale, assuming
                    # panel is 800 pixels wide
                    scalebar = AnchoredSizeBar(
                        ax.transData,
                        800 / (width_ * Rh.d),
                        "1 Mpc",
                        "lower right",
                        pad=0.4,
                        label_top=False,
                        sep=10,
                        color=text_color,
                        frameon=False,
                        size_vertical=1,
                    )
                    ax.add_artist(scalebar)

            halo_id_previous = halo_id

    return fig



from enum import Enum
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Concatenate, TypeVar

import h5py

from opencosmo.mpi import get_comm_world

H5Resource = TypeVar("H5Resource", h5py.File, h5py.Group, h5py.Dataset)
H5Reader = Callable[Concatenate[H5Resource, ...], Any]
FileReader = Callable[Concatenate[h5py.File | h5py.Group, ...], Any]
FileWriter = Callable[Concatenate[h5py.File | h5py.Group, ...], None]

"""
A collection of utilities for working with hdf5 files.
"""


class FileExistance(Enum):
    MUST_EXIST = "must_exist"
    MUST_NOT_EXIST = "must_not_exist"
    EITHER = "either"


def file_reader(func: FileReader) -> FileReader:
    """
    Resolves the path to a given file and handles opening
    and closing the file for reading purposes. Used as a
    decorator.
    """

    @wraps(func)
    def wrapper(file: h5py.File | h5py.Group | Path | str, *args, **kwargs):
        if not isinstance(file, (h5py.File, h5py.Group)):
            path = resolve_path(file, FileExistance.MUST_EXIST)
            with h5py.File(path, "r", driver="core") as f:
                return func(f, *args, **kwargs)
        return func(file, *args, **kwargs)

    return wrapper


def file_writer(func: FileWriter) -> FileWriter:
    """
    Resolves the path to a given file and handles opening
    and closing the file for writing purposes. Used as a
    decorator.
    """

    @wraps(func)
    def wrapper(file: h5py.File | Path | str, *args, **kwargs):
        if not isinstance(file, h5py.File | h5py.Group):
            file = resolve_path(file, FileExistance.MUST_NOT_EXIST)
        return func(file, *args, **kwargs)

    return wrapper


def broadcast_read(func: H5Reader) -> H5Reader:
    """
    If MPI is available, the decorated function will only
    be called on rank 0, with the results broadcast to all
    ranks. When we read actual data this is not necessary,
    because the data will be chunked by rank, but for attributes
    its better to not try to read them from a hundred processes
    at the same time.
    """

    @wraps(func)
    def wrapper(file: h5py.File | Path | str, *args, **kwargs):
        output = None
        comm = get_comm_world()
        if comm is None or comm.Get_rank() == 0:
            try:
                output = func(file, *args, **kwargs)
            except Exception as e:
                output = e
        if comm is not None:
            # Broadcasting the error ensures other ranks
            # will raise the exception and quit.
            output = comm.bcast(output, root=0)
        if isinstance(output, Exception):
            raise output
        return output

    return wrapper


def get_data_structure(group: h5py.Group) -> dict[str, Any]:
    units = {name: group[name].attrs.get("unit", "") for name in group.keys()}
    return units


def resolve_path(
    path: Path | str, existance: FileExistance = FileExistance.MUST_EXIST
) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if not path.exists() and existance == FileExistance.MUST_EXIST:
        raise FileNotFoundError(f"{path} does not exist.")
    if path.exists() and existance == FileExistance.MUST_NOT_EXIST:
        raise FileExistsError(f"{path} already exists.")
    if not path.parent.exists():
        raise FileNotFoundError(f"Directory {path.parent} does not exist!")
    if path.suffix != ".hdf5":
        raise ValueError(f"{path} does not appear to be an hdf5 file.")
    return path



from __future__ import annotations

from functools import cache, partial
from typing import TYPE_CHECKING, Optional

import astropy.units as u
from astropy.cosmology import units as cu

from opencosmo.units import UnitConvention

if TYPE_CHECKING:
    import numpy as np
    from astropy.cosmology import Cosmology

    from opencosmo.dataset.state import DatasetState

KNOWN_SCALEFACTOR_COLUMNS = {"fof_halo_center_a"}
KNOWN_REDSHIFT_COLUMNS = {"redshift", "redshift_true"}


def get_unit_transitions(
    unit: u.Unit,
    base_convention: UnitConvention,
    cosmology: Cosmology,
    is_comoving: bool,
):
    match base_convention:
        case (UnitConvention.PHYSICAL, UnitConvention.UNITLESS):
            return {}, {}
        case UnitConvention.SCALEFREE:
            return get_scalefree_transitions(unit, cosmology, is_comoving)
        case UnitConvention.COMOVING:
            return get_comoving_transitions(unit, cosmology, is_comoving)


def get_comoving_transitions(unit: u.Unit, cosmology: Cosmology, is_comoving: bool):
    error = partial(
        raise_convert_error, from_=UnitConvention.COMOVING, to_=UnitConvention.SCALEFREE
    )

    transitions = {UnitConvention.SCALEFREE: error}
    inv_transitions = {UnitConvention.SCALEFREE: error}
    distance_power = get_unit_distance_power
    if distance_power is not None and is_comoving:
        transitions[UnitConvention.PHYSICAL] = comoving_to_physical  # type: ignore
        inv_transitions[UnitConvention.PHYSICAL] = partial(
            physical_to_comoving,
            base_unit=unit,
        )
    units = {UnitConvention.COMOVING: unit, UnitConvention.PHYSICAL: unit}
    return transitions, inv_transitions, units


def get_scalefree_transitions(unit: u.Unit, cosmology: Cosmology, is_comoving: bool):
    hless_unit = get_unit_without_h(unit)
    transitions = {}
    inv_transitions = {}

    if hless_unit != unit:
        rem_h = partial(remove_littleh, cosmology=cosmology)
        add_h = partial(add_littleh, cosmology=cosmology, new_unit=unit)
        transitions[UnitConvention.COMOVING] = rem_h
        inv_transitions[UnitConvention.COMOVING] = add_h

    distance_power = get_unit_distance_power(unit)
    if distance_power is not None and is_comoving:
        transitions[UnitConvention.PHYSICAL] = partial(
            scalefree_to_physical, cosmology=cosmology
        )
        inv_transitions[UnitConvention.PHYSICAL] = partial(
            physical_to_scalefree, base_unit=unit, cosmology=cosmology
        )
    elif transitions.get(UnitConvention.COMOVING) is not None:
        transitions[UnitConvention.PHYSICAL] = transitions[UnitConvention.COMOVING]
        inv_transitions[UnitConvention.PHYSICAL] = inv_transitions[
            UnitConvention.COMOVING
        ]

    units = {
        UnitConvention.SCALEFREE: unit,
        UnitConvention.COMOVING: hless_unit,
        UnitConvention.PHYSICAL: hless_unit,
    }

    return transitions, inv_transitions, units


@cache
def get_unit_without_h(unit: u.Unit) -> u.Unit:
    try:
        if isinstance(unit, u.DexUnit):
            u_base = unit.physical_unit
            constructor = u.DexUnit
        else:
            u_base = unit

            def constructor(x):
                return x

    except AttributeError:
        return unit

    try:
        index = u_base.bases.index(cu.littleh)
    except (ValueError, AttributeError):
        return unit
    power = u_base.powers[index]
    new_unit = u_base / cu.littleh**power

    return new_unit


@cache
def get_unit_distance_power(unit: u.Unit) -> Optional[float]:
    decomposed = unit.decompose()
    try:
        index = decomposed.bases.index(u.m)
        return decomposed.powers[index]
    except (ValueError, AttributeError):
        return None


def add_littleh(value: u.Quantity, cosmology: Cosmology, new_unit: u.Unit, **kwargs):
    return value.to(new_unit, cu.with_H0(cosmology.H0))


def physical_to_comoving(
    value: u.Quantity, scale_factor: float | np.ndarray, base_unit: u.Unit, **kwargs
):
    power = get_unit_distance_power(base_unit)
    if power is not None:
        return value / (scale_factor**power)
    return value


def remove_littleh(value: u.Quantity, cosmology: Cosmology, **kwargs) -> u.Quantity:
    """
    Remove little h from the units of the input table. For comoving
    coordinates, this is the second step after parsing the units themselves.
    """
    new_unit = get_unit_without_h(value.unit)
    if new_unit != value.unit:
        return value.to(new_unit, cu.with_H0(cosmology.H0))
    return value


def comoving_to_physical(
    value: u.Quantity, scale_factor: float | np.ndarray, **kwargs
) -> u.Quantity:
    """
    Convert comoving coordinates to physical coordinates. This is the
    second step after parsing the units themselves.
    """

    unit = value.unit
    # Check if the units have distances in them
    power = get_unit_distance_power(unit)
    # multiply by the scale factor to the same power as the distance
    if power is not None:
        return value * scale_factor**power
    return value


def get_scale_factor(dataset: "DatasetState", cosmology, redshift):
    columns = set(dataset.columns)
    for column in KNOWN_SCALEFACTOR_COLUMNS:
        if column in columns:
            col = dataset.select(column).get_data()[column].data
            return col

    for column in KNOWN_REDSHIFT_COLUMNS:
        if column in columns:
            col = dataset.select(column).get_data()[column].data
            return 1 / (1 + col)

    return cosmology.scale_factor(redshift)


def scalefree_to_physical(
    value: u.Quantity,
    scale_factor: float | np.ndarray,
    cosmology: Cosmology,
):
    new_value = remove_littleh(value, cosmology)
    return comoving_to_physical(new_value, scale_factor)


def physical_to_scalefree(
    value: u.Quantity,
    scale_factor: float | np.ndarray,
    base_unit: u.Quantity,
    cosmology: Cosmology,
):
    new_value = physical_to_comoving(value, scale_factor, base_unit)
    return add_littleh(new_value, cosmology, base_unit)


def raise_convert_error(*args, from_: UnitConvention, to_: UnitConvention, **kwargs):
    raise ValueError(
        f"Units in convention {str(from_)} cannot be converted to units in convention {str(to_)}"
    )



from __future__ import annotations

from itertools import count
from typing import TYPE_CHECKING, Sequence
from uuid import uuid1

import h5py
import numpy as np

try:
    from mpi4py import MPI
except ImportError:
    MPI = None  # type: ignore

from opencosmo.index import ChunkedIndex, SimpleIndex
from opencosmo.io.schemas import (
    ColumnSchema,
)
from opencosmo.spatial.healpix import HealPixIndex
from opencosmo.spatial.octree import OctTreeIndex
from opencosmo.spatial.protocols import TreePartition
from opencosmo.spatial.utils import combine_upwards

if TYPE_CHECKING:
    from opencosmo.index import DataIndex
    from opencosmo.spatial.protocols import Region, SpatialIndex


def open_tree(file: h5py.File | h5py.Group, box_size: int, is_lightcone: bool = False):
    """
    Read a tree from an HDF5 file and the associated
    header. The tree is just a mapping between a spatial
    index and a slice into the data.

    Note: The max level in the header may not actually match
    the max level in the file. When a large dataset is filtered down,
    we may reduce the tree level to save space in the output file.

    The max level in the header is the maximum level in the full
    dataset, so this is the HIGHEST it can be.
    """
    try:
        group = file["index"]
    except KeyError:
        raise ValueError("This file does not have a spatial index!")

    if is_lightcone:
        spatial_index = HealPixIndex()
    else:
        spatial_index = OctTreeIndex.from_box_size(box_size)
    return Tree(spatial_index, group)


def read_tree(file: h5py.File | h5py.Group, box_size: int):
    try:
        group = file["index"]
    except KeyError:
        raise ValueError("This file does not have a spatial index!")

    f = h5py.File(f"{uuid1()}.hdf5", "w", driver="core", backing_store=False)
    for ds in group.keys():
        group.copy(ds, f)
    spatial_index = OctTreeIndex.from_box_size(box_size)
    return Tree(spatial_index, f)


def apply_range_mask(
    mask: np.ndarray,
    range_: tuple[int, int],
    starts: dict[int, np.ndarray],
    sizes: dict[int, np.ndarray],
) -> dict[int, tuple[int, np.ndarray]]:
    """
    Given an index range, apply a mask of the same size to produces new sizes.
    """
    output_sizes = {}

    for level, st in starts.items():
        ends = st + sizes[level]
        # Not in range if the end is less than start, or the start is greater than end
        overlaps_mask = ~((st > range_[1]) | (ends < range_[0]))
        # The first start may be less thank the range start so
        first_start_index = int(np.argmax(overlaps_mask))
        st = st[overlaps_mask]
        st[0] = range_[0]
        st = st - range_[0]
        # Determine how many true values are in the mask in the ranges
        new_sizes = np.add.reduceat(mask, st)
        output_sizes[level] = (first_start_index, new_sizes)
    return output_sizes


def pack_masked_ranges(
    old_starts: dict[int, np.ndarray],
    new_sizes: list[dict[int, tuple[int, np.ndarray]]],
    min_level_size: int = 500,
) -> tuple[dict[int, np.ndarray], dict[int, np.ndarray]]:
    """
    Given a list of masked ranges, pack them into a new set of sizes.
    This is used when working with MPI, and allows us to avoid sending
    very large masks between ranks.

    For queries that return a small fraction of the data, we can end up
    writing a lot of zeros in the lower levels of the tree. So we can
    dynamically choose to stop writing levels when the average size of
    the level is below a certain threshold
    """
    output_starts = {}
    output_sizes = {}
    for level in new_sizes[0]:
        new_level_sizes = np.zeros_like(old_starts[level])
        new_start_info = [rm[level] for rm in new_sizes]
        for first_idx, sizes in new_start_info:
            new_level_sizes[first_idx : first_idx + len(sizes)] += sizes

        avg_size = np.mean(new_level_sizes[new_level_sizes > 0])
        if avg_size < min_level_size:
            break
        output_sizes[level] = new_level_sizes
        output_starts[level] = np.cumsum(np.insert(new_level_sizes, 0, 0))[:-1]

    return output_starts, output_sizes


def partition_index(n_partitions: int, counts: h5py.Group):
    levels = [int(key.split("_")[1]) for key in counts.keys()]
    lowest_level = min(levels)
    highest_level = max(levels)
    split_level = -1
    for level in range(lowest_level, highest_level + 1):
        level_counts = counts[f"level_{level}"]["size"][:]
        full_region_indices = np.where(level_counts > 0)[0]
        n_full = len(full_region_indices)
        if n_full < n_partitions:
            continue
        elif n_full % n_partitions == 0:
            split_level = level
            break

    if split_level == -1:
        split_level = highest_level

    split_level_indices = full_region_indices

    partition_indices = np.array_split(split_level_indices, n_partitions)
    return [SimpleIndex(idx) for idx in partition_indices], split_level


class Tree:
    """
    The Tree handles the spatial indexing of the data.
    """

    def __init__(self, index: SpatialIndex, data: h5py.File | h5py.Group):
        self.__index = index
        self.__data = data
        for i in count():
            try:
                _ = self.__data[f"level_{i}"]["start"]
                _ = self.__data[f"level_{i}"]["size"]
            except KeyError:
                self.__max_level = i - 1
                break

        if self.__max_level == -1:
            raise ValueError("Tried to read a tree but no levels were found!")

    def partition(
        self, n_partitions: int, counts: h5py.Group
    ) -> Sequence[TreePartition]:
        """
        Partition into n trees, where each tree contains an equally sized
        region of space.

        This function is used primarily in an MPI context.
        """
        partition_indices, split_level = partition_index(n_partitions, counts)
        partitions = []
        start = self.__data[f"level_{split_level}"]["start"]
        size = self.__data[f"level_{split_level}"]["size"]
        for index_ in partition_indices:
            if len(index_) == 0:
                continue
            index_starts = index_.get_data(start)
            index_sizes = index_.get_data(size)
            partition_start = index_starts[0]
            partition_size = np.sum(index_sizes)
            idx = ChunkedIndex.single_chunk(partition_start, partition_size)
            region = self.__index.get_partition_region(index_, split_level)
            partitions.append(TreePartition(idx, region, split_level))

        return partitions

    def query(self, region: Region) -> tuple[ChunkedIndex, ChunkedIndex]:
        indices = self.__index.query(region, self.__max_level)

        contains = [ChunkedIndex.empty()]
        intersects = [ChunkedIndex.empty()]
        for level, (cidx, iidx) in indices.items():
            level_key = f"level_{level}"
            level_starts = self.__data[level_key]["start"]
            level_sizes = self.__data[level_key]["size"]
            c_starts = cidx.get_data(level_starts)
            c_sizes = cidx.get_data(level_sizes)
            i_starts = iidx.get_data(level_starts)
            i_sizes = iidx.get_data(level_sizes)
            c_idx = ChunkedIndex(c_starts, c_sizes)
            i_idx = ChunkedIndex(i_starts, i_sizes)
            contains.append(c_idx)
            intersects.append(i_idx)

        c_ = contains[0].concatenate(*contains[1:])
        i_ = intersects[0].concatenate(*intersects[1:])
        return c_, i_

    def apply_index(self, index: DataIndex, min_counts: int = 100) -> Tree:
        max_level_starts = self.__data[f"level_{self.__max_level}"]["start"][:]
        max_level_sizes = self.__data[f"level_{self.__max_level}"]["size"][:]
        n_in_range = index.n_in_range(max_level_starts, max_level_sizes)
        target = h5py.File(f"{uuid1()}.hdf5", "w", driver="core", backing_store=False)
        result = combine_upwards(
            n_in_range, self.__index.subdivision_factor, self.__max_level, target
        )
        return Tree(self.__index, result)

    def make_schema(self):
        columns = {}
        for level in range(self.__max_level + 1):
            source = self.__data[f"level_{level}"]
            index = ChunkedIndex.from_size(len(source["start"]))
            start = ColumnSchema(
                f"level_{level}/start", index, source["start"], source["start"].attrs
            )
            size = ColumnSchema(
                f"level_{level}/size", index, source["size"], source["size"].attrs
            )

            columns[f"index/level_{level}/start"] = start
            columns[f"index/level_{level}/size"] = size

        return columns



from __future__ import annotations

from enum import Enum
from functools import reduce
from typing import TYPE_CHECKING, Callable, Iterable, Optional

import h5py
from deprecated import deprecated  # type: ignore

import opencosmo as oc
from opencosmo import collection
from opencosmo.dataset import state as dss
from opencosmo.dataset.handler import Hdf5Handler
from opencosmo.dataset.im import InMemoryColumnHandler
from opencosmo.file import FileExistance, file_reader, resolve_path
from opencosmo.header import read_header
from opencosmo.index import ChunkedIndex
from opencosmo.mpi import get_comm_world
from opencosmo.spatial.builders import from_model
from opencosmo.spatial.region import FullSkyRegion
from opencosmo.spatial.tree import open_tree, read_tree
from opencosmo.units import UnitConvention
from opencosmo.units.get import get_unit_applicators_hdf5
from opencosmo.units.handler import make_unit_handler

from .schemas import FileSchema

if TYPE_CHECKING:
    from pathlib import Path
    from types import ModuleType

    from opencosmo.header import OpenCosmoHeader

    from .protocols import Writeable

mpiio: Optional[ModuleType]
partition: Optional[Callable]

if get_comm_world() is not None:
    from opencosmo.dataset.mpi import partition
    from opencosmo.io import mpi as mpiio
else:
    mpiio = None
    partition = None

"""
This module defines the main user-facing io functions: open and write

open can take any number of file paths, and will always construct a single object 
(either a dataset or a collection).

write takes exactly one path and exactly one opencosmo dataset or collection

open works in the following way:

1. Read headers and get dataset names and types for all files passed
2. If there is only a single dataset, simply open it as such
3. If there are multiple datasets, user the headers to determine
   if the dataset are compatible (i.e. capabale of existing together in
   a collection)
4. Open all datasets individually
5. Call the merge functionality for the appropriate collection.
"""


class FILE_TYPE(Enum):
    HALO_PROPERTIES = 0
    HALO_PARTICLES = 1
    GALAXY_PROPERTIES = 2
    GALAXY_PARTICLES = 3
    SOD_BINS = 4
    LIGHTCONE = 5
    STRUCTURE_COLLECTION = 6
    SIMULATION_COLLECTION = 7
    SYNTHETIC_CATALOG = 8


class COLLECTION_TYPE(Enum):
    LIGHTCONE = 0
    STRUCTURE_COLLECTION = 1
    SIMULATION_COLLECTION = 2


class OpenTarget:
    def __init__(self, group: h5py.Group | h5py.File, header: OpenCosmoHeader):
        self.group = group
        self.header = header

    @property
    def data_type(self):
        return self.header.file.data_type


def get_file_type(file: h5py.File) -> FILE_TYPE:
    if "header" in file.keys():
        dtype = file["header"]["file"].attrs["data_type"]
        if dtype == "halo_particles":
            return FILE_TYPE.HALO_PARTICLES
        elif dtype == "halo_profiles":
            return FILE_TYPE.SOD_BINS
        elif dtype == "halo_properties":
            return FILE_TYPE.HALO_PROPERTIES
        elif dtype == "galaxy_properties":
            return FILE_TYPE.GALAXY_PROPERTIES
        elif dtype == "galaxy_particles":
            return FILE_TYPE.GALAXY_PARTICLES
        elif dtype == "diffsky_fits":
            return FILE_TYPE.SYNTHETIC_CATALOG
        else:
            raise ValueError(f"Unknown file type {dtype}")

    if not all("header" in group.keys() for group in file.values()):
        for subgroup in file.values():
            if not all("header" in g.keys() for g in subgroup.values()):
                raise ValueError(
                    "Unknown file type. "
                    "It appears to have multiple datasets, but organized incorrectly"
                )
    if all(group["header"]["file"].attrs["is_lightcone"] for group in file.values()):
        return FILE_TYPE.LIGHTCONE
    elif (
        len(set(group["header"]["file"].attrs["data_type"] for group in file.values()))
        == 1
    ):
        return FILE_TYPE.SIMULATION_COLLECTION

    elif all("data" not in group.keys() for group in file.values()):
        for group in file.values():
            sub_groups = {
                g["header"]["file"].attrs["data_type"]: g for g in group.values()
            }
            collection.structure.io.validate_linked_groups(sub_groups)
        return FILE_TYPE.SIMULATION_COLLECTION
    else:
        group = {name: group for name, group in file.items()}
        collection.structure.io.validate_linked_groups(group)
        return FILE_TYPE.STRUCTURE_COLLECTION


def make_all_targets(files: list[h5py.File]):
    targets: list[OpenTarget] = reduce(
        lambda targets, file: targets + make_file_targets(file), files, []
    )
    return targets


def make_file_targets(file: h5py.File):
    try:
        header = read_header(file, unit_convention=UnitConvention.COMOVING)
    except KeyError:
        header = None
    if header is not None and "data" in file.keys():
        return [OpenTarget(file, header)]
    if header is None and "data" in file.keys():
        raise ValueError(
            "This file appears to be missing a header. "
            "Are you sure it is an OpenCosmo file?"
        )
    if header is None:
        headers = {name: read_header(group) for name, group in file.items()}
    else:
        headers = {name: header for name in file.keys() if name != "header"}

    output = []
    for name, header in headers.items():
        target = OpenTarget(file[name], header)
        output.append(target)
    return output


def open(
    *files: str | Path | h5py.File | h5py.Group, **open_kwargs: bool
) -> oc.Dataset | collection.Collection:
    """
    Open a dataset or data collection from one or more opencosmo files.

    If you open a file with this function, you should generally close it
    when you're done

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("path/to/file.hdf5")
        # do work
        ds.close()

    Alternatively you can use a context manager, which will close the file
    automatically when you are done with it.

    .. code-block:: python

        import opencosmo as oc
        with oc.open("path/to/file.hdf5") as ds:
            # do work

    When you have multiple files that can be combined into a collection,
    you can use the following.

    .. code-block:: python

        import opencosmo as oc
        ds = oc.open("haloproperties.hdf5", "haloparticles.hdf5")


    Parameters
    ----------
    *files: str or pathlib.Path
        The path(s) to the file(s) to open.

    **open_kwargs: bool
        True/False flags that can be used to only load certain datasets from
        the files. Check the documentation for the data type you are working
        with for available flags. Will be ignored if only one file is passed
        and the file only contains a single dataset.

    Returns
    -------
    dataset : oc.Dataset or oc.Collection
        The dataset or collection opened from the file.

    """
    if len(files) == 1 and isinstance(files[0], list):
        file_list = files[0]
    else:
        file_list = list(files)
    file_list.sort()
    handles = [h5py.File(f) for f in file_list]
    file_types = list(map(get_file_type, handles))
    targets = make_all_targets(handles)
    targets = evaluate_load_conditions(targets, open_kwargs)
    if len(targets) > 1:
        collection_type = collection.get_collection_type(targets, file_types)
        return collection_type.open(targets, **open_kwargs)

    else:
        return open_single_dataset(targets[0])

    # For now the only way to open multiple files is with a StructureCollection


def open_single_dataset(
    target: OpenTarget,
    metadata_group: Optional[str] = None,
    bypass_lightcone: bool = False,
    bypass_mpi: bool = False,
):
    header = target.header
    handle = target.group

    assert header is not None

    try:
        tree = open_tree(
            handle,
            header.with_units("scalefree").simulation["box_size"].value,
            header.file.is_lightcone,
        )
    except ValueError:
        tree = None

    if header.file.region is not None:
        sim_region = from_model(header.file.region)
    elif header.file.is_lightcone:
        sim_region = FullSkyRegion()
    else:
        p1 = (0, 0, 0)
        p2 = tuple(header.simulation["box_size"].value for _ in range(3))
        sim_region = oc.make_box(p1, p2)

    index: Optional[ChunkedIndex] = None
    handler = Hdf5Handler.from_group(handle["data"])

    if not bypass_mpi and (comm := get_comm_world()) is not None:
        assert partition is not None
        idx_data = handle["index"]

        part = partition(comm, len(handler), idx_data, tree)
        if part is None:
            index = ChunkedIndex.empty()
        else:
            index = part.idx
            sim_region = part.region if part.region is not None else sim_region

    if metadata_group is not None:
        metadata_group = handle[metadata_group]

    state = dss.DatasetState.from_group(
        handle["data"],
        header,
        UnitConvention.COMOVING,
        sim_region,
        index,
        metadata_group,
    )

    dataset = oc.Dataset(
        header,
        state,
        tree=tree,
    )

    if header.file.is_lightcone and not bypass_lightcone:
        return collection.Lightcone({"data": dataset}, header.lightcone["z_range"])

    return dataset


def get_file_handles(*files: str | Path | h5py.File | h5py.Group):
    handles = []
    for file in files:
        if not isinstance(file, h5py.File) and not isinstance(file, h5py.Group):
            path = resolve_path(file, FileExistance.MUST_EXIST)
            file_handle = h5py.File(path, "r")
            handles.append(file_handle)
            continue
        handles.append(file)
    return handles


def evaluate_load_conditions(targets: list[OpenTarget], open_kwargs: dict[str, bool]):
    """
    Datasets can define conditional loading via an addition group called "load/if".
    the "if" group can define parameters which must either be true or false for the
    given group to be loaded. These parameters can then be provided by the user to the
    "open" function. Parameters not specified by the user default to False.

    Note that some open kwargs may be used in other places in the opening process,
    and will just be ignored here.
    """
    if len(targets) == 1:
        return targets
    output = []
    for target in targets:
        try:
            ifgroup = target.group["load/if"]
        except KeyError:
            output.append(target)
            continue
        load = True
        for key, condition in ifgroup.attrs.items():
            load = load and (open_kwargs.get(key, False) == condition)
        if load:
            output.append(target)
    return output


def write(path: Path, dataset: Writeable, overwrite=False) -> None:
    """
    Write a dataset or collection to the file at the sepecified path.

    Parameters
    ----------
    file : str or pathlib.Path
        The path to the file to write to.
    dataset : oc.Dataset
        The dataset to write.
    overwrite : bool, default = False
        If the file already exists, overwrite it


    Raises
    ------
    FileExistsError
        If the file at the specified path already exists and overwrite is False
    FileNotFoundError
        If the parent folder of the ouput file does not exist
    """

    existance_requirement = FileExistance.MUST_NOT_EXIST
    if overwrite:
        existance_requirement = FileExistance.EITHER

    path = resolve_path(path, existance_requirement)

    schema = FileSchema()
    dataset_schema = dataset.make_schema()
    schema.add_child(dataset_schema, "root")

    if mpiio is not None:
        return mpiio.write_parallel(path, schema)

    file = h5py.File(path, "w")
    schema.allocate(file)

    writer = schema.into_writer()

    writer.write(file)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

import astropy.units as u  # type: ignore
import numpy as np
from astropy.coordinates import SkyCoord  # type: ignore

if TYPE_CHECKING:
    from opencosmo.dataset.dataset import Dataset
    from opencosmo.parameters import FileParameters
    from opencosmo.spatial.protocols import Region

ALLOWED_COORDINATES_3D = {
    "default": {
        "fof": "fof_halo_center_",
        "mass": "fof_halo_com_",
        "sod": "sod_halo_com_",
    }
}


def check_containment(
    ds: "Dataset",
    region: "Region",
    parameters: FileParameters,
    select_by: Optional[str] = None,
):
    dtype = str(parameters.data_type)
    if parameters.is_lightcone:
        return __check_containment_2d(ds, region, dtype)
    else:
        return __check_containment_3d(ds, region, dtype)


def get_theta_phi_coordinates(dataset: "Dataset"):
    coord_values = dataset.select(["theta", "phi"]).data
    ra = coord_values["phi"]
    dec = np.pi / 2 - coord_values["theta"]

    return SkyCoord(ra, dec, unit=u.rad)


def find_coordinates_2d(dataset: "Dataset"):
    columns = set(dataset.columns)
    if len(columns.intersection(set(["theta", "phi"]))) == 2:
        return get_theta_phi_coordinates(dataset)
    elif len(columns.intersection(set(["ra", "dec"]))) == 2:
        data = dataset.select(["ra", "dec"]).data
        return SkyCoord(data["ra"], data["dec"])
    raise ValueError("Dataset does not contain coordinates")


def find_coordinates_3d(ds: "Dataset", dtype: str, select_by: Optional[str] = None):
    try:
        allowed_coordinates = ALLOWED_COORDINATES_3D[dtype]
    except KeyError:
        allowed_coordinates = ALLOWED_COORDINATES_3D["default"]
    if select_by is None:
        column_name_base = next(iter(allowed_coordinates.values()))
    else:
        column_name_base = allowed_coordinates[dtype]

    cols = set(filter(lambda colname: colname.startswith(column_name_base), ds.columns))
    expected_cols = [column_name_base + dim for dim in ["x", "y", "z"]]
    if cols != set(expected_cols):
        raise ValueError(
            "Unable to find the correct coordinate columns in this dataset! "
            f"Found {cols} but expected {expected_cols}"
        )
    return expected_cols


def __check_containment_3d(
    ds: "Dataset", region: "Region", dtype: str, select_by: Optional[str] = None
):
    columns = find_coordinates_3d(ds, dtype, select_by)
    ds = ds.select(columns)
    data = ds.data

    data = np.vstack(tuple(data[col].data for col in columns))
    return region.contains(data)


def __check_containment_2d(
    ds: "Dataset", region: "Region", dtype: str, select_by: Optional[str] = None
):
    coords = find_coordinates_2d(ds)
    return region.contains(coords)



from .io import open, write


def missing_package_import_error(name, required):
    raise ValueError(f"Function {name} requires package {required}")


__all__ = ["write", "open", "write_parquet"]


def __getattr__(name):
    if name == "write_parquet":
        try:
            from .parquet import write_parquet
        except ImportError:
            raise ImportError("To use write_parquet you must install pyarrow")
        return write_parquet
    raise ImportError(f"cannot import name '{name}' from 'opencosmo.io'")
